"""AI REST Commands"""

from typing import Final, TypeGuard

from async_reolink.api.commands import ai

# from async_reolink.api.ai import typings

from ..ai import models
from . import (
    _CHANNEL_KEY,
    CommandResponse,
    CommandResponseTypes,
    CommandRequestWithChannel,
    CommandResponseWithChannel,
)

# pylint:disable=missing-function-docstring


class GetAiStateRequest(CommandRequestWithChannel, ai.GetAiStateRequest):
    """Get AI State"""

    __slots__ = ()

    COMMAND: Final = "GetAiState"

    def __init__(
        self,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id


_NONE_DICT: Final[dict] = None


class GetAiStateResponse(
    CommandResponse,
    ai.GetAiStateResponse,
    test="is_response",
):
    """Get AI State Response"""

    __slots__ = ()

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetAiStateRequest.COMMAND)

    @property
    def channel_id(self) -> int:
        if (value := self._get_value()) is None:
            return None
        return value.get(_CHANNEL_KEY, None)

    @property
    def state(self):
        return models.State(self._get_value())


class GetAiConfigRequest(CommandRequestWithChannel, ai.GetAiConfigRequest):
    """Get AI Configuration"""

    COMMAND: Final = "GetAiCfg"

    def __init__(
        self,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id


class GetAiConfigResponse(
    CommandResponseWithChannel,
    ai.GetAiConfigResponse,
    test="is_response",
):
    """Get AI Configuration Response"""

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetAiConfigRequest.COMMAND)

    @property
    def channel_id(self) -> int:
        if (value := self._get_value()) is None:
            return None
        return value.get(_CHANNEL_KEY, None)

    @property
    def config(self):
        return models.Config(self._get_value())


class SetAiConfigRequest(CommandRequestWithChannel, ai.SetAiConfigRequest):
    """Set AI Configuration"""

    COMMAND: Final = "SetAiCfg"

    def __init__(
        self,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ) -> None:
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id

    @property
    def config(self):
        return models.MutableConfig(self._get_parameter)



"""LED Models"""

from typing import Callable, Final
from async_reolink.api.led import typings
from async_reolink.api.typings import SimpleTimeValue

from ..typings import FactoryValue

from ..ai.models import AITypesMap, MutableAITypesMap

# pylint: disable=missing-function-docstring


class LightingSchedule(typings.LightingSchedule):
    """Lighting Schedule"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        super().__init__()
        self._factory = factory

    class StartTime(SimpleTimeValue):
        """Start Time"""

        _HOUR_KEY: Final = "StartHour"
        _MIN_KEY: Final = "StartMin"

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        @property
        def hour(self):
            if (value := self._factory()) is None:
                return 0
            return value.get(type(self)._HOUR_KEY, 0)

        @property
        def minute(self):
            if (value := self._factory()) is None:
                return 0
            return value.get(type(self)._MIN_KEY, 0)

    @property
    def start(self):
        return type(self).StartTime(self._factory)

    class EndTime(SimpleTimeValue):
        """End Time"""

        __slots__ = ("_factory",)

        _HOUR_KEY: Final = "EndHour"
        _MIN_KEY: Final = "EndMin"

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        @property
        def hour(self):
            if (value := self._factory()) is None:
                return 0
            return value.get(type(self)._HOUR_KEY, 0)

        @property
        def minute(self):
            if (value := self._factory()) is None:
                return 0
            return value.get(type(self)._MIN_KEY, 0)

    @property
    def end(self):
        return type(self).EndTime(self._factory)


class MutableLightingSchedule(LightingSchedule):
    """Mutable Lighting schedule"""

    def __init__(self, factory: FactoryValue[dict]) -> None:
        super().__init__(factory)
        self._factory = factory

    class StartTime(LightingSchedule.StartTime):
        """Start Time"""

        def __init__(self, factory: FactoryValue[dict]) -> None:
            super().__init__(factory)
            self._factory = factory

        @LightingSchedule.StartTime.hour.setter
        def hour(self, value):
            if (value := self._factory(True)) is not None:
                value[type(self)._HOUR_KEY] = value

        @LightingSchedule.StartTime.minute.setter
        def minute(self, value):
            if (value := self._factory(True)) is not None:
                value[type(self)._MIN_KEY] = value

    def _update_time(self, _time: SimpleTimeValue, value: SimpleTimeValue):
        _time.hour = value.hour
        _time.minute = value.minute

    @LightingSchedule.start.setter
    def start(self, value):
        self._update_time(self.start, value)

    class EndTime(LightingSchedule.EndTime):
        """End Time"""

        def __init__(self, factory: FactoryValue[dict]) -> None:
            super().__init__(factory)
            self._factory = factory

        @LightingSchedule.EndTime.hour.setter
        def hour(self, value):
            if (value := self._factory(True)) is not None:
                value[type(self)._HOUR_KEY] = value

        @LightingSchedule.EndTime.minute.setter
        def minute(self, value):
            if (value := self._factory(True)) is not None:
                value[type(self)._MIN_KEY] = value

    @LightingSchedule.end.setter
    def end(self, value: SimpleTimeValue):
        self._update_time(self.end, value)


class WhiteLedInfo(typings.WhiteLedInfo):
    """White Led Info"""

    _BRIGHT_KEY: Final = "bright"
    _AUTO_KEY: Final = "auto"
    _MODE_KEY: Final = "mode"
    _STATE_KEY: Final = "state"
    _SCHED_KEY: Final = "LightingSchedule"
    _AI_KEY: Final = "wlAiDetectType"

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        super().__init__()
        self._factory = factory

    @property
    def brightness(self):
        if (value := self._factory()) is None:
            return 100
        return value.get(type(self)._BRIGHT_KEY, 100)

    @property
    def auto_mode(self):
        if (value := self._factory()) is None:
            return 0
        return value.get(type(self)._AUTO_KEY, 0)

    @property
    def brightness_state(self):
        if (value := self._factory()) is None:
            return 0
        return value.get(type(self)._MODE_KEY, 0)

    @property
    def state(self):
        if (value := self._factory()) is None:
            return 0
        return value.get(type(self)._STATE_KEY, 0)

    def _get_sched(self) -> dict:
        if (value := self._factory()) is None:
            return None
        return value.get(type(self)._SCHED_KEY, None)

    @property
    def lighting_schedule(self):
        return LightingSchedule(self._get_sched)

    def _get_ai(self) -> dict:
        if (value := self._factory()) is None:
            return None
        return value.get(type(self)._AI_KEY, None)

    @property
    def ai_detection_type(self):
        return AITypesMap(self._get_ai)


class MutableWhiteLedInfo(WhiteLedInfo):
    """White Led Info"""

    def __init__(self, factory: FactoryValue[dict]) -> None:
        super().__init__(factory)
        self._factory = factory

    @WhiteLedInfo.brightness.setter
    def brightness(self, value):
        if (value := self._factory(True)) is not None:
            value[type(self)._BRIGHT_KEY] = value

    @WhiteLedInfo.auto_mode.setter
    def auto_mode(self, value):
        if (value := self._factory(True)) is not None:
            value[type(self)._AUTO_KEY] = value

    @WhiteLedInfo.brightness_state.setter
    def brightness_state(self, value):
        if (value := self._factory(True)) is not None:
            value[type(self)._MODE_KEY] = value

    @WhiteLedInfo.state.setter
    def state(self, value):
        if (value := self._factory(True)) is not None:
            value[type(self)._STATE_KEY] = value

    def _get_sched(self, create=False) -> dict:
        if not create:
            return super()._get_sched()
        if (value := self._factory(True)) is None:
            return None
        return value.setdefault(type(self)._SCHED_KEY, {})

    @WhiteLedInfo.lighting_schedule.getter
    def lighting_schedule(self):
        return MutableLightingSchedule(self._get_sched)

    @lighting_schedule.setter
    def lighting_schedule(self, value: typings.LightingSchedule):
        _sched = self.lighting_schedule
        _sched.start = value.start
        _sched.end = value.end

    def _get_ai(self, create=False) -> dict:
        if not create:
            return super()._get_ai()
        if (value := self._factory(True)) is None:
            return None
        return value.setdefault(type(self)._AI_KEY, {})

    @WhiteLedInfo.ai_detection_type.getter
    def ai_detection_type(self):
        return MutableAITypesMap(self._get_ai)

    @ai_detection_type.setter
    def ai_detection_type(self, value):
        _type: MutableAITypesMap = self.ai_detection_type
        _type.update(value)



"""REST PTZ Mixin"""

from async_reolink.api import ptz
from async_reolink.api.ptz.typings import (
    Patrol,
    Preset,
    ZoomOperation,
    Operation,
    Track,
)

from async_reolink.rest.ptz.models import MutablePatrol

from ..commands import ptz as commands


class PTZ(ptz.PTZ):
    """REST PTZ Mixin"""

    def _create_get_ptz_autofocus_request(self, channel: int):
        return commands.GetAutoFocusRequest(channel)

    def _create_set_ptz_autofocus_request(self, channel: int, disabled: bool):
        return commands.SetAutoFocusRequest(disabled, channel_id=channel)

    def _create_get_ptz_patrols_request(self, channel: int):
        return commands.GetPatrolRequest(channel)

    def _create_set_ptz_patrol_request(self, patrol: Patrol):
        return commands.SetPatrolRequest(patrol)

    def _create_get_ptz_presets_request(self, channel: int):
        return commands.GetPresetRequest(channel)

    def _create_set_ptz_preset_request(self, preset: Preset):
        return commands.SetPresetRequest(preset)

    def _create_get_ptz_tatterns_request(self, channel: int):
        return commands.GetTatternRequest(channel)

    def _create_set_ptz_tatterns_request(self, channel: int, *track: Track):
        return commands.SetTatternRequest(*track, channel_id=channel)

    def _create_get_ptz_zoom_focus_request(self, channel: int):
        return commands.GetZoomFocusRequest(channel)

    def _create_set_ptz_zoomfocus_request(
        self, channel: int, operation: ZoomOperation, position: int
    ):
        return commands.SetZoomFocusRequest(operation, position, channel)

    def _create_set_ptz_control_request(
        self,
        channel: int,
        operation: Operation,
        speed: int | None,
        preset_id: int | None,
    ):
        return commands.SetControlRequest(operation, speed, preset_id, channel)



"""Encoding REST Commands"""

from typing import Final, TypeGuard
from async_reolink.api.commands import encoding

from ..encoding.models import EncodingInfo

from ..commands import (
    _CHANNEL_KEY,
    CommandResponseTypes,
    CommandRequestWithChannel,
    CommandResponse,
)

# pylint: disable=missing-function-docstring


class GetEncodingRequest(CommandRequestWithChannel, encoding.GetEncodingRequest):
    """Get Encoding REST Request"""

    COMMAND: Final = "GetEnc"

    def __init__(
        self,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id


class GetEncodingResponse(
    CommandResponse, encoding.GetEncodingResponse, test="is_response"
):
    """Get Encoding REST Response"""

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetEncodingRequest.COMMAND)

    def _get_info(self) -> dict:
        if (value := self._get_value()) is None:
            return None
        return value.get("Enc", None)

    @property
    def channel_id(self) -> int:
        if (info := self._get_info()) is not None:
            return info.get(_CHANNEL_KEY, 0)
        return 0

    @property
    def info(self):
        return EncodingInfo(self._get_info)



"""REST Commands"""

from abc import ABC
from enum import IntEnum
from json import dumps
from typing import (
    Callable,
    Final,
    TypeGuard,
)

from async_reolink.api import commands

_COMMAND_KEY: Final = "cmd"
_ACTION_KEY: Final = "action"

# pylint: disable=missing-function-docstring


class CommandResponseTypes(IntEnum):
    """Command Response Types"""

    VALUE_ONLY = 0
    DETAILED = 1


class CommandRequest(commands.CommandRequest):
    """Rest Command Request"""

    __slots__ = ("_request",)

    def __init__(self):
        self._request = {}

    def _get_request(self):
        return self._request

    def _get_parameter(self, create=False) -> dict:
        _key: Final = "param"

        if _key in self._request or not create:
            return self._request.get(_key, None)
        return self._request.setdefault(_key, {})

    @property
    def _parameter(self):
        return self._get_parameter(True)

    @property
    def command(self) -> str:
        return self._request.get(_COMMAND_KEY, "")

    @command.setter
    def command(self, value):
        self._request[_COMMAND_KEY] = value

    @property
    def response_type(self) -> CommandResponseTypes:
        return self._request.get(_ACTION_KEY, CommandResponseTypes.VALUE_ONLY)

    @response_type.setter
    def response_type(self, value):
        self._request[_ACTION_KEY] = value


_CHANNEL_KEY: Final = "channel"


class CommandRequestWithChannel(CommandRequest, commands.ChannelValue):
    """Rest Command Request with Channel Parameter"""

    __slots__ = ()

    @property
    def channel_id(self):
        return (
            parameter.get(_CHANNEL_KEY, 0)
            if (parameter := self._get_parameter()) is not None
            else 0
        )

    @channel_id.setter
    def channel_id(self, value):
        self._parameter[_CHANNEL_KEY] = value


_VALUE_KEY: Final = "value"
_CODE_KEY: Final = "code"
_ERROR_KEY: Final = "error"


class CommandResponse(commands.CommandResponse, ABC):
    """Rest Command Response"""

    _response_handlers: dict[type["CommandResponse"], Callable[[dict], bool]] = {}

    def __init_subclass__(  # pylint: disable=arguments-differ
        cls, test: str = None, **kwargs
    ) -> None:
        super().__init_subclass__(**kwargs)
        if (
            isinstance(test, str)
            and (call := getattr(cls, test, None)) is not None
            and callable(call)
        ):
            cls._response_handlers[cls] = call

    @classmethod
    def create_from(cls, value: dict):
        """wrap response in CommandResponse Implementation"""

        for (_type, test) in cls._response_handlers.items():
            if test(value):
                return _type(value)

        return CommandResponse(value)

    @classmethod
    def is_response(cls, value: any, /, command: str | None = None) -> TypeGuard[dict]:
        """value is command response json"""
        return (
            isinstance(value, dict)
            and _COMMAND_KEY in value
            and _CODE_KEY in value
            and (command is None or command == value[_COMMAND_KEY])
        )

    @classmethod
    def is_value(cls, response: dict):
        """response is a command value response"""
        return _VALUE_KEY in response and isinstance(response[_VALUE_KEY], dict)

    __slots__ = ("_response",)

    def __init__(self, response: dict) -> None:
        self._response = response

    def _underlying_value(self):
        return self._response

    @property
    def command(self) -> str:
        return self._response.get(_COMMAND_KEY, "")

    @property
    def code(self) -> int:
        return self._response.get(_CODE_KEY, 0)

    def _get_value(self) -> dict:
        return self._response.get(_VALUE_KEY, None)

    def _get_initial(self) -> dict:
        return self._response.get("initial", None)

    def _get_range(self) -> dict:
        return self._response.get("range", None)

    def _get_error(self) -> dict:
        return self._response.get(_ERROR_KEY, None)


class UnhandledCommandResponse(CommandResponse):
    """Unhandled/Unknown REST Command Response"""

    __slots__ = ()


_RSP_CODE_KEY: Final = "rspCode"


class CommandResponseWithCode(
    CommandResponse, commands.ResponseCode, test="is_response"
):
    """REST Command Response with code"""

    @classmethod
    def is_response(  # pylint: disable=signature-differs
        cls, value: any, /
    ) -> TypeGuard["CommandResponseWithCode"]:
        return (
            super().is_response(value)
            and super().is_value(value)
            and _RSP_CODE_KEY in value[_VALUE_KEY]
        )

    __slots__ = ()

    @property
    def response_code(self):
        return (
            value.get(_RSP_CODE_KEY, 0)
            if (value := self._get_value()) is not None
            else 0
        )


class CommandResponseWithChannel(CommandResponse, commands.ChannelValue):
    """Rest Command Response Value with Channel"""

    __slots__ = ()

    @property
    def channel_id(self):
        return (
            value.get(_CHANNEL_KEY, 0)
            if (value := self._get_value()) is not None
            else 0
        )


class CommandErrorResponse(
    CommandResponse, commands.CommandErrorResponse, test="is_error"
):
    """Rest Command Error Response"""

    __slots__ = ()

    @classmethod
    def is_error(cls, response: dict):
        """response is a command error response"""
        return _ERROR_KEY in response and isinstance(response[_ERROR_KEY], dict)

    @property
    def error_code(self) -> int:
        return (
            value.get(_RSP_CODE_KEY, 0)
            if (value := self._get_error()) is not None
            else 0
        )

    @property
    def details(self) -> str | None:
        return value.get("detail", 0) if (value := self._get_error()) is not None else 0



"""System Mixin"""

from async_reolink.api import system

from ..commands import system as commands
from .capabilities import Capabilities


class System(system.System):
    """Rest System Mixin"""

    def _create_get_capabilities_request(self, username: str | None):
        return commands.GetAbilitiesRequest(
            username if username is not None else "null"
        )

    def _create_empty_capabilities(self):
        def _factory():
            return None

        return Capabilities(_factory)

    def _create_get_device_info_request(self):
        return commands.GetDeviceInfoRequest()

    def _create_empty_device_info(self):
        def _factory():
            return None

        return commands.DeviceInfo(_factory)

    def _create_get_time_request(self):
        return commands.GetTimeRequest()

    def _create_reboot_request(self):
        return commands.RebootRequest()



"""REST Record Models"""

from datetime import date, datetime, time
from typing import Callable, Final, Iterable, overload
from async_reolink.api.typings import DateTimeValue as BaseDateTimeValue, StreamTypes
from async_reolink.api.record import typings

from ..typings import STR_STREAMTYPES_MAP, STREAMTYPES_STR_MAP, FactoryValue

# pylint: disable=missing-function-docstring


class DateTimeValue(BaseDateTimeValue):
    """REST Recording DateTime Value"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        self._factory = factory

    @property
    def year(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("year", 0)

    @property
    def month(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("mon", 0)

    @property
    def day(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("day", 0)

    def to_date(self):
        return date(self.year, self.month, self.day)

    @property
    def hour(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("hour", 0)

    @property
    def minute(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("min", 0)

    @property
    def second(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("sec", 0)

    def to_time(self):
        return time(self.hour, self.minute, self.second)

    def to_datetime(self):
        return datetime(self.to_date(), self.to_time())


class MutableDateTimeValue(DateTimeValue):
    """REST Recording Mutable DateTime Value"""

    __slots__ = ()

    @overload
    def __init__(self) -> None:
        ...

    @overload
    def __init__(self, value: datetime) -> None:
        ...

    @overload
    def __init__(self, value: typings.DateTimeValue) -> None:
        ...

    @overload
    def __init__(self, factory: FactoryValue[dict]) -> None:
        ...

    def __init__(self, factory: FactoryValue[dict] = None) -> None:
        source: typings.DateTimeValue = None
        if not callable(factory):
            value = factory
            if not isinstance(value, dict):
                source = value
                value = {}

            def _factory(*_):
                return value

            factory = _factory
        super().__init__(factory)
        self._factory = factory
        if source is not None:
            self.year = source.year
            self.month = source.month
            self.day = source.day
            self.hour = source.hour
            self.minute = source.minute
            self.second = source.second

    @property
    def _value(self):
        return self._factory(True)

    @DateTimeValue.year.setter
    def year(self, value):
        self._value["year"] = value

    @DateTimeValue.month.setter
    def month(self, value):
        self._value["mon"] = value

    @DateTimeValue.day.setter
    def day(self, value):
        self._value["day"] = value

    @DateTimeValue.hour.setter
    def hour(self, value):
        self._value["hour"] = value

    @DateTimeValue.minute.setter
    def minute(self, value):
        self._value["min"] = value

    @DateTimeValue.second.setter
    def second(self, value):
        self._value["sec"] = value


_DEFAULT_STREAMTYPE: Final = StreamTypes.MAIN
_DEFAULT_STREAMTYPE_STR = STREAMTYPES_STR_MAP[_DEFAULT_STREAMTYPE]


class Search(typings.Search):
    """REST Search"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        self._factory = factory

    @property
    def status_only(self) -> bool:
        if (value := self._factory()) is None:
            return 0
        return value.get("onlyStatus", 0)

    @property
    def stream_type(self):
        if (value := self._factory()) is None:
            return _DEFAULT_STREAMTYPE
        return STR_STREAMTYPES_MAP[value.get("streamType", _DEFAULT_STREAMTYPE_STR)]

    def _get_start(self) -> dict:
        if (value := self._factory()) is None:
            return None
        return value.get("StartTime", None)

    @property
    def start(self):
        return DateTimeValue(self._get_start)

    def _get_end(self) -> dict:
        if (value := self._factory()) is None:
            return None
        return value.get("EndTime", None)

    @property
    def end(self):
        return DateTimeValue(self._get_end)


class MutableSearch(Search):
    """REST Mutable Search"""

    __slots__ = ()

    @overload
    def __init__(self) -> None:
        ...

    @overload
    def __init__(self, value: typings.Search) -> None:
        ...

    @overload
    def __init__(self, factory: FactoryValue[dict]) -> None:
        ...

    def __init__(self, factory: FactoryValue[dict] = None) -> None:
        source: typings.Search = None
        if not callable(factory):
            value = factory
            if not isinstance(value, dict):
                source = value
                value = {}

            def _factory(*_):
                return value

            factory = _factory
        super().__init__(factory)
        self._factory = factory
        if source is not None:
            self.status_only = source.status_only
            self.stream_type = source.stream_type
            self.start = source.start
            self.end = source.end

    @property
    def _value(self):
        return self._factory(True)

    @Search.status_only.setter
    def status_only(self, value):
        self._value["onlyStatus"] = int(bool(value))

    @Search.stream_type.setter
    def stream_type(self, value):
        self._value["streamType"] = STREAMTYPES_STR_MAP[value or _DEFAULT_STREAMTYPE]

    def _get_start(self, create=False) -> dict:
        _key: Final = "StartTime"
        if (value := self._factory(create)) is None:
            return None
        if _key in value or not create:
            return value.get(_key, None)
        return value.setdefault(_key, {})

    @property
    def start(self):
        return MutableDateTimeValue(self._get_start)

    @start.setter
    def start(self, value: typings.DateTimeValue):
        if not isinstance(value, MutableDateTimeValue):
            value = MutableDateTimeValue(value)
        self._value["StartTime"] = value._factory(True)

    def _get_end(self, create=False) -> dict:
        _key: Final = "EndTime"
        if (value := self._factory(create)) is None:
            return None
        if _key in value or not create:
            return value.get(_key, None)
        return value.setdefault(_key, {})

    @property
    def end(self):
        return MutableDateTimeValue(self._get_end)

    @end.setter
    def end(self, value: typings.DateTimeValue):
        if not isinstance(value, MutableDateTimeValue):
            value = MutableDateTimeValue(value)
        self._value["EndTime"] = value._factory(True)


class File(typings.File):
    """REST Recording File"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        self._factory = factory

    def _get_start(self) -> dict:
        if (value := self._factory()) is None:
            return None
        return value.get("StartTime", None)

    @property
    def frame_rate(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("frameRate", 0)

    @property
    def width(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("width", 0)

    @property
    def height(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("height", 0)

    @property
    def size(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("size", 0)

    @property
    def name(self) -> str:
        if (value := self._factory()) is None:
            return 0
        return value.get("name", 0)

    @property
    def type(self) -> str:
        if (value := self._factory()) is None:
            return 0
        return value.get("type", 0)

    @property
    def start(self):
        return DateTimeValue(self._get_start)

    def _get_end(self) -> dict:
        if (value := self._factory()) is None:
            return None
        return value.get("EndTime", None)

    @property
    def end(self):
        return DateTimeValue(self._get_end)


class _SearchStatusTable(Iterable[int]):
    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        self._factory = factory

    def __iter__(self):
        if (value := self._factory()) is None:
            return
        for i, _c in enumerate(str(value), 1):
            if _c == "1":
                yield i


class SearchStatus(typings.SearchStatus):
    """REST Recodring Search Status"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        self._factory = factory

    @property
    def year(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("year", 0)

    @property
    def month(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("mon", 0)

    def _get_table(self) -> str:
        if (value := self._factory()) is None:
            return ""
        return value.get("table", "")

    @property
    def days(self):
        return _SearchStatusTable(self._get_table)

    def __iter__(self):
        year = self.year
        month = self.month
        for _d in self.days:
            yield date(year, month, _d)



"""REST PTZ Typings"""

from types import MappingProxyType
from typing import Final
from async_reolink.api.ptz.typings import ZoomOperation, Operation

STR_OPERATION_MAP: Final = MappingProxyType(
    {
        "Stop": Operation.STOP,
        "Left": Operation.LEFT,
        "Right": Operation.RIGHT,
        "Up": Operation.UP,
        "LeftUp": Operation.LEFT_UP,
        "LeftDown": Operation.LEFT_DOWN,
        "RightUp": Operation.RIGHT_UP,
        "RightDown": Operation.RIGHT_DOWN,
        "IrisDec": Operation.IRIS_SHRINK,
        "IrisInc": Operation.IRIS_ENLARGE,
        "ZoomDec": Operation.ZOOM_OUT,
        "ZoomInc": Operation.ZOOM_IN,
        "FocusInc": Operation.FOCUS_BACK,
        "FocusDec": Operation.FOCUS_FORWARD,
        "Auto": Operation.AUTO,
        "StartPatrol": Operation.PATROL_START,
        "StopPatrol": Operation.PATROL_STOP,
        "ToPos": Operation.TO_PRESET,
    }
)

OPERATION_STR_MAP: Final = MappingProxyType(
    {_v: _k for _k, _v in STR_OPERATION_MAP.items()}
)

STR_ZOOMOPERATION_MAP: Final = MappingProxyType(
    {"ZoomPos": ZoomOperation.ZOOM, "FocusPos": ZoomOperation.FOCUS}
)

ZOOMOPERATION_STR_MAP: Final = MappingProxyType(
    {_v: _k for _k, _v in STR_ZOOMOPERATION_MAP.items()}
)



"""REST Alarm Commands"""

from typing import Final, TypeGuard
from async_reolink.api.commands import alarm

from . import CommandRequestWithChannel, CommandResponse, CommandResponseTypes


class GetMotionStateRequest(CommandRequestWithChannel, alarm.GetMotionStateRequest):
    """REST Get Motion State Request"""

    COMMAND: Final = "GetMdState"

    def __init__(
        self,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id


class GetMotionStateResponse(
    CommandResponse, alarm.GetMostionStateResponse, test="is_response"
):
    """REST Get Motion State Response"""

    __slots__ = ()

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetMotionStateRequest.COMMAND)

    @property
    def state(self) -> bool:
        return value.get("state", 0) if (value := self._get_value()) is not None else 0



"""AI Mixin"""

from async_reolink.api import encoding

from ..commands import encoding as commands


class Encoding(encoding.Encoding):
    """Rest Encoding Mixin"""

    def _create_get_encoding_request(self, channel: int):
        return commands.GetEncodingRequest(channel)



"""REST Network Typings"""

from types import MappingProxyType
from typing import Final

from async_reolink.api.network.typings import LinkTypes

STR_LINKTYPES_MAP: Final = MappingProxyType(
    {"Static" if _e == LinkTypes.STATIC else _e.name.upper(): _e for _e in LinkTypes}
)

LINKTYPES_STR_MAP: Final = MappingProxyType(
    {_v: _k for _k, _v in STR_LINKTYPES_MAP.items()}
)



"""REST PTZ Models"""

from typing import Callable, Final, Iterable, MutableSequence, Sequence, overload
from async_reolink.api.ptz import typings

from ..typings import FactoryValue

from ..commands import _CHANNEL_KEY

# pylint: disable=missing-function-docstring

_ID_KEY: Final = "id"
_ENABLE_KEY: Final = "enable"
_NAME_KEY: Final = "name"

_FOCUS_KEY: Final = "focus"
_ZOOM_KEY: Final = "zoom"


class ZoomFocus(typings.ZoomFocus):
    """REST PTZ Zoom/Focus"""

    __slots__ = ("_value",)

    def __init__(self, value: dict) -> None:
        super().__init__()
        if value is None:
            value = {}
        self._value = value

    @property
    def zoom(self) -> int:
        return self._value.get(_ZOOM_KEY, 0)

    @property
    def focus(self) -> int:
        return self._value.get(_FOCUS_KEY, 0)


class Preset(typings.Preset):
    """REST PTZ Preset"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        self._factory = factory

    @property
    def channel_id(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get(_CHANNEL_KEY, 0)

    @property
    def id(self) -> int:  # pylint: disable=invalid-name
        if (value := self._factory()) is None:
            return 0
        return value.get(_ID_KEY, 0)

    @property
    def enabled(self) -> bool:
        if (value := self._factory()) is None:
            return 0
        return value.get(_ENABLE_KEY, 0)

    @property
    def name(self) -> str:
        if (value := self._factory()) is None:
            return ""
        return value.get(_NAME_KEY, "")


class MutablePreset(Preset):
    """Mutable PTZ Preset"""

    __slots__ = ()

    @overload
    def __init__(self) -> None:
        ...

    @overload
    def __init__(self, value: typings.Preset) -> None:
        ...

    @overload
    def __init__(self, factory: FactoryValue[dict]) -> None:
        ...

    def __init__(self, factory: FactoryValue[dict] = None) -> None:
        source: typings.Preset = None
        if not callable(factory):
            value = factory
            if not isinstance(value, dict):
                source = value
                value = {}

            def _factory(*_):
                return value

            factory = _factory
        super().__init__(factory)
        self._factory = factory
        if source is not None:
            self.channel_id = source.channel_id
            self.enabled = source.enabled
            self.id = source.id
            self.name = source.name

    @Preset.channel_id.setter
    def channel_id(self, value):
        if (value := self._factory(True)) is not None:
            value[_CHANNEL_KEY] = value

    @Preset.id.setter
    def id(self, value):
        if (value := self._factory(True)) is not None:
            value[_ID_KEY] = value

    @Preset.name.setter
    def name(self, value):
        if (value := self._factory(True)) is not None:
            value[_NAME_KEY] = value


_DWELL_TIME_KEY: Final = "dwellTime"
_SPEED_KEY: Final = "speed"


class PatrolPreset(typings.PatrolPreset):
    """REST PTZ Patrol Preset"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        self._factory = factory

    @property
    def dwell_time(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get(_DWELL_TIME_KEY, 0)

    @property
    def preset_id(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get(_ID_KEY, 0)

    @property
    def speed(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get(_SPEED_KEY, 0)


class MutablePatrolPreset(PatrolPreset):
    """REST Mutable PTZ Patrol Preset"""

    __slots__ = ()

    @overload
    def __init__(self) -> None:
        ...

    @overload
    def __init__(self, value: typings.PatrolPreset) -> None:
        ...

    @overload
    def __init__(self, factory: FactoryValue[dict]) -> None:
        ...

    def __init__(
        self,
        factory: FactoryValue[dict] = None,
    ) -> None:
        source: typings.Preset = None
        if not callable(factory):
            value = factory
            if not isinstance(value, dict):
                source = value
                value = {}

            def _factory(*_):
                return value

            factory = _factory
        super().__init__(factory)
        self._factory = factory
        if source is not None:
            self.dwell_time = source.dwell_time
            self.preset_id = source.preset_id
            self.speed = source.speed

    @PatrolPreset.dwell_time.setter
    def dwell_time(self, value):
        if (value := self._factory(True)) is not None:
            value[_DWELL_TIME_KEY] = value

    preset_id = Preset.id

    @PatrolPreset.speed.setter
    def speed(self, value):
        if (value := self._factory(True)) is not None:
            value[_SPEED_KEY] = value


class _PatrolPresets(Sequence[PatrolPreset]):
    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], list]) -> None:
        self._factory = factory

    def __getitem__(self, __k: int):
        def _factory():
            if (value := self._factory()) is None:
                return None
            return value[__k]

        return PatrolPreset(_factory)

    def __len__(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return len(value)


class _MutablePatrolPresets(MutableSequence[MutablePatrolPreset]):
    __slots__ = ("_factory",)

    def __init__(self, factory: FactoryValue[list]) -> None:
        self._factory = factory

    def __getitem__(self, __k: int):
        def _factory():
            if (value := self._factory()) is None:
                return None
            return value[__k]

        return MutablePatrolPreset(_factory)

    def __setitem__(self, __k: int, __v: MutablePatrolPreset):
        if (value := self._factory(True)) is not None:
            value[__k] = __v._factory(True)

    def __delitem__(self, __k: int):
        if (value := self._factory()) is not None:
            del value[__k]

    def _insert(self, index: int, value: dict):
        if (_value := self._factory(True)) is not None:
            _value.insert(index, value)

    def append(self, value: typings.PatrolPreset) -> None:
        return super().append(value)

    def insert(self, index: int, value: typings.PatrolPreset):
        if not isinstance(value, MutablePatrolPreset):
            value = MutablePatrolPreset(value)
        # pylint: disable=protected-access
        self._insert(index, value._factory(True))

    def clear(self) -> None:
        if (value := self._factory()) is not None:
            value.clear()

    def __len__(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return len(value)


_PRESET_KEY: Final = "preset"
_RUNNING_KEY: Final = "running"


class Patrol(typings.Patrol):
    """REST PTZ Patrol"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        self._factory = factory

    channel_id = Preset.channel_id
    id = Preset.id
    enabled = Preset.enabled
    name = Preset.name

    def _get_presets(self) -> list:
        if (value := self._factory()) is None:
            return None
        return value.get(_PRESET_KEY, None)

    @property
    def presets(self):
        return _PatrolPresets(self._get_presets)

    @property
    def running(self) -> bool:
        if (value := self._factory()) is None:
            return 0
        return value.get(_RUNNING_KEY, 0)


class MutablePatrol(Patrol):
    """Mutable REST PTZ Patrol"""

    __slots__ = ()

    @overload
    def __init__(self) -> None:
        ...

    @overload
    def __init__(self, value: typings.Patrol) -> None:
        ...

    @overload
    def __init__(self, factory: FactoryValue[dict]) -> None:
        ...

    def __init__(self, factory: FactoryValue[dict] = None) -> None:
        source: typings.Patrol = None
        if not callable(factory):
            value = factory
            if not isinstance(value, dict):
                source = value
                value = {}

            def _factory(*_):
                return value

            factory = _factory
        super().__init__(factory)
        self._factory = factory
        if source is not None:
            self.channel_id = source.channel_id
            self.enabled = source.enabled
            self.id = source.id
            self.name = source.name
            self.presets = source.presets

    channel_id = MutablePreset.channel_id
    id = MutablePreset.id
    enabled = MutablePreset.enabled
    name = MutablePreset.name

    def _get_presets(self, create=False) -> list:
        _key: Final = _PRESET_KEY
        if (value := self._factory(create)) is None:
            return None
        if _key in value or not create:
            return value.get(_key, None)
        return value.setdefault(_key, {})

    @property
    def presets(self):
        return _MutablePatrolPresets(self._get_presets)

    @presets.setter
    def presets(self, value: Iterable[typings.PatrolPreset]):
        _presets = self.presets
        _presets.clear()
        for preset in value:
            _presets.append(preset)


class Track(typings.Track):
    """REST Track (Tattern)"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        self._factory = factory

    id = Preset.id
    enabled = Preset.enabled
    name = Preset.name
    running = Patrol.running


class MutableTrack(Track):
    """Mutable REST Track (Tattern)"""

    __slots__ = ()

    @overload
    def __init__(self) -> None:
        ...

    @overload
    def __init__(self, value: typings.Track) -> None:
        ...

    @overload
    def __init__(self, factory: FactoryValue[dict]) -> None:
        ...

    def __init__(self, factory: FactoryValue[dict] = None) -> None:
        source: typings.Track = None
        if not callable(factory):
            value = factory
            if not isinstance(value, dict):
                source = value
                value = {}

            def _factory(*_):
                return value

            factory = _factory
        super().__init__(factory)
        self._factory = factory
        if source is not None:
            self.enabled = source.enabled
            self.id = source.id
            self.name = source.name
            self.running = source.running

    id = MutablePreset.id
    enabled = MutablePreset.enabled
    name = MutablePreset.name
    running = MutablePatrol.running



"""System capabilities"""

from typing import Callable, Mapping, TypeVar, overload
from async_reolink.api.system import capabilities

_T = TypeVar("_T")

# pylint: disable=too-few-public-methods
# pylint: disable=missing-function-docstring

_defaults: dict[type, any] = {None: 0}


class Capability(capabilities.Capability[_T]):
    """Capability"""

    __slots__ = ("_factory", "_init")

    @overload
    def __init__(self: "Capability[int]", factory: Callable[[], any] = None):
        ...

    @overload
    def __init__(
        self, factory: Callable[[], dict], __type: Callable[[int], _T]
    ) -> None:
        ...

    def __init__(
        self, factory: Callable[[], dict], __type: Callable[[int], _T] = None
    ) -> None:
        super().__init__()
        self._factory = factory
        self._init = __type
        if __type not in _defaults:
            _defaults[__type] = __type(0)

    def _get_value(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("ver", 0)

    @property
    def value(self) -> _T:
        value = self._get_value()
        if value == 0:
            return _defaults[self._init]
        if self._init is not None:
            return self._init(value)
        return value

    @property
    def permissions(self) -> capabilities.Permissions:
        if (value := self._factory()) is None:
            return 0
        return value.get("perm", 0)

    def __bool__(self):
        return bool(self._get_value())

    def __index__(self):
        return self._get_value()

    def __int__(self):
        return self._get_value()

    def __str__(self):
        return str(self.value)

    def __eq__(self, __o: object) -> bool:
        if hasattr(__o, "__index__"):
            return self._get_value() == __o.__index__()
        return self._get_value() == __o

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._factory())}>"


class ChannelCapabilities(capabilities.ChannelCapabilities):
    """Channel Capabilities"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        super().__init__()
        self._factory = factory

    def _keyed_factory(self, key: str):
        def _factory():
            if (value := self._factory()) is None:
                return None
            return value.get(key, None)

        return _factory

    class AI(capabilities.ChannelCapabilities.AI):
        """AI"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        class Track(capabilities.ChannelCapabilities.AI.Track):
            """Track"""

            __slots__ = ("_factory",)

            def __init__(self, factory: Callable[[], dict]) -> None:
                super().__init__()
                self._factory = factory

            def _keyed_factory(self, key: str):
                def _factory():
                    if (value := self._factory()) is None:
                        return None
                    return value.get(key, None)

                return _factory

            @property
            def _value(self):
                return Capability(self._keyed_factory("aiTrack"))

            @property
            def value(self):
                return self._value.value

            @property
            def permissions(self):
                return self._value.permissions

            def __bool__(self):
                return bool(self.value)

            @property
            def pet(self):
                return Capability(self._keyed_factory("aiTrackDogCat"))

        @property
        def track(self):
            return type(self).Track(self._factory)

    @property
    def ai(self):
        return type(self).AI(self._factory)

    class Alarm(capabilities.ChannelCapabilities.Alarm):
        """Alarm"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        def _keyed_factory(self, key: str):
            def _factory():
                if (value := self._factory()) is None:
                    return None
                return value.get(key, None)

            return _factory

        @property
        def audio(self):
            return Capability(self._keyed_factory("alarmAudio"))

        @property
        def io_in(self):
            return Capability(self._keyed_factory("alarmIoIn"))

        @property
        def io_out(self):
            return Capability(self._keyed_factory("alarmIoOut"))

        @property
        def motion(self):
            return Capability(self._keyed_factory("alarmMd"))

        @property
        def rf(self):
            return Capability(self._keyed_factory("alarmRf"))

    @property
    def alarm(self):
        return type(self).Alarm(self._factory)

    @property
    def battery(self):
        return Capability(self._keyed_factory("battery"))

    @property
    def battery_analysis(self):
        return Capability(self._keyed_factory("batAnalysis"))

    @property
    def camera_mode(self):
        return Capability(self._keyed_factory("cameraMode"))

    @property
    def disable_autofocus(self):
        return Capability(self._keyed_factory("disableAutoFocus"))

    @property
    def enc(self):
        return Capability(self._keyed_factory("enc"))

    @property
    def floodlight(self):
        return Capability(self._keyed_factory("floodLight"))

    @property
    def ftp(self):
        return Capability(self._keyed_factory("ftp"))

    @property
    def image(self):
        return Capability(self._keyed_factory("image"))

    @property
    def indicator_light(self):
        return Capability(self._keyed_factory("indicatorLight"))

    class ISP(capabilities.ChannelCapabilities.ISP):
        """ISP"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        def _keyed_factory(self, key: str):
            def _factory():
                if (value := self._factory()) is None:
                    return None
                return value.get(key, None)

            return _factory

        @property
        def _value(self):
            return Capability(self._keyed_factory("isp"))

        @property
        def value(self):
            return self._value.value

        @property
        def permissions(self):
            return self._value.permissions

        def __bool__(self):
            return bool(self.value)

        @property
        def threeDnr(self):
            return Capability(self._keyed_factory("isp3Dnr"))

        @property
        def antiflicker(self):
            return Capability(self._keyed_factory("ispAntiFlick"))

        @property
        def backlight(self):
            return Capability(self._keyed_factory("ispBackLight"))

        @property
        def bright(self):
            return Capability(self._keyed_factory("ispBright"))

        @property
        def contrast(self):
            return Capability(self._keyed_factory("ispContrast"))

        @property
        def day_night(self):
            return Capability(self._keyed_factory("ispDayNight"))

        @property
        def exposure_mode(self):
            return Capability(self._keyed_factory("ispExposureMode"))

        @property
        def flip(self):
            return Capability(self._keyed_factory("ispFlip"))

        @property
        def hue(self):
            return Capability(self._keyed_factory("ispHue"))

        @property
        def mirror(self):
            return Capability(self._keyed_factory("ispMirror"))

        @property
        def satruation(self):
            return Capability(self._keyed_factory("ispSatruation"))

        @property
        def sharpen(self):
            return Capability(self._keyed_factory("ispSharpen"))

        @property
        def white_balance(self):
            return Capability(self._keyed_factory("ispWhiteBalance"))

    @property
    def isp(self):
        return type(self).ISP(self._factory)

    @property
    def led_control(self):
        return Capability(self._keyed_factory("ledControl"))

    @property
    def live(self):
        return Capability(self._keyed_factory("live"))

    @property
    def main_encoding(self):
        return Capability(self._keyed_factory("mainEncType"))

    @property
    def mask(self):
        return Capability(self._keyed_factory("mask"))

    class MD(capabilities.ChannelCapabilities.MD):
        """MotionDetection"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        def _keyed_factory(self, key: str):
            def _factory():
                if (value := self._factory()) is None:
                    return None
                return value.get(key, None)

            return _factory

        class Trigger(capabilities.ChannelCapabilities.MD.Trigger):
            """Trigger"""

            __slots__ = ("_factory",)

            def __init__(self, factory: Callable[[], dict]) -> None:
                super().__init__()
                self._factory = factory

            def _keyed_factory(self, key: str):
                def _factory():
                    if (value := self._factory()) is None:
                        return None
                    return value.get(key, None)

                return _factory

            @property
            def audio(self):
                return Capability(self._keyed_factory("mdTriggerAudio"))

            @property
            def record(self):
                return Capability(self._keyed_factory("mdTriggerRecord"))

        @property
        def trigger(self):
            return type(self).Trigger(self._factory)

        @property
        def with_pir(self):
            return Capability(self._keyed_factory("mdWithPir"))

    @property
    def motion_detection(self):
        return type(self).MD(self._factory)

    @property
    def osd(self):
        return Capability(self._keyed_factory("osd"))

    @property
    def power_led(self):
        return Capability(self._keyed_factory("powerLed"))

    class PTZ(capabilities.ChannelCapabilities.PTZ):
        """PTZ"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        def _keyed_factory(self, key: str):
            def _factory():
                if (value := self._factory()) is None:
                    return None
                return value.get(key, None)

            return _factory

        @property
        def control(self):
            return Capability(self._keyed_factory("ptzCtrl"))

        @property
        def direction(self):
            return Capability(self._keyed_factory("ptzDirection"))

        @property
        def patrol(self):
            return Capability(self._keyed_factory("ptzPatrol"))

        @property
        def preset(self):
            return Capability(self._keyed_factory("ptzPreset"))

        @property
        def tattern(self):
            return Capability(self._keyed_factory("ptzTattern"))

        @property
        def type(self):
            return Capability(self._keyed_factory("ptzType"))

    @property
    def ptz(self):
        return type(self).PTZ(self._factory)

    class Record(capabilities.ChannelCapabilities.Record):
        """Record"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        def _keyed_factory(self, key: str):
            def _factory():
                if (value := self._factory()) is None:
                    return None
                return value.get(key, None)

            return _factory

        @property
        def config(self):
            return Capability(self._keyed_factory("recCfg"))

        @property
        def download(self):
            return Capability(self._keyed_factory("recDownload"))

        @property
        def replay(self):
            return Capability(self._keyed_factory("recReplay"))

        @property
        def schedule(self):
            return Capability(self._keyed_factory("recSchedule"))

    @property
    def record(self):
        return type(self).Record(self._factory)

    @property
    def shelter_config(self):
        return Capability(self._keyed_factory("shelterCfg"))

    @property
    def snap(self):
        return Capability(self._keyed_factory("snap"))

    class Supports(capabilities.ChannelCapabilities.Supports):
        """Supports"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        def _keyed_factory(self, key: str):
            def _factory():
                if (value := self._factory()) is None:
                    return None
                return value.get(key, None)

            return _factory

        class AI(capabilities.ChannelCapabilities.Supports.AI):
            """AI"""

            __slots__ = ("_factory",)

            def __init__(self, factory: Callable[[], dict]) -> None:
                super().__init__()
                self._factory = factory

            def _keyed_factory(self, key: str):
                def _factory():
                    if (value := self._factory()) is None:
                        return None
                    return value.get(key, None)

                return _factory

            @property
            def _value(self):
                return Capability(self._keyed_factory("supportAi"))

            @property
            def value(self):
                return self._value.value

            @property
            def permissions(self):
                return self._value.permissions

            def __bool__(self):
                return bool(self.value)

            @property
            def animal(self):
                return Capability(self._keyed_factory("supportAiAnimal"))

            @property
            def detect_config(self):
                return Capability(self._keyed_factory("supportAiDetectConfig"))

            @property
            def pet(self):
                return Capability(self._keyed_factory("supportAiDogCat"))

            @property
            def face(self):
                return Capability(self._keyed_factory("supportAiFace"))

            @property
            def people(self):
                return Capability(self._keyed_factory("supportAiPeople"))

            @property
            def sensitivity(self):
                return Capability(self._keyed_factory("supportAiSensitivity"))

            @property
            def stay_time(self):
                return Capability(self._keyed_factory("supportAiStayTime"))

            @property
            def target_size(self):
                return Capability(self._keyed_factory("supportAiTargetSize"))

            @property
            def track_classify(self):
                return Capability(self._keyed_factory("supportAiTrackClassify"))

            @property
            def vehicle(self):
                return Capability(self._keyed_factory("supportAiVehicle"))

            @property
            def adjust(self):
                return Capability(self._keyed_factory("supportAoAdjust"))

        @property
        def ai(self):
            return type(self).AI(self._factory)

        class FloodLight(capabilities.ChannelCapabilities.Supports.FloodLight):
            """FloodLight"""

            __slots__ = ("_factory",)

            def __init__(self, factory: Callable[[], dict]) -> None:
                super().__init__()
                self._factory = factory

            def _keyed_factory(self, key: str):
                def _factory():
                    if (value := self._factory()) is None:
                        return None
                    return value.get(key, None)

                return _factory

            @property
            def brightness(self):
                return Capability(self._keyed_factory("supportFLBrightness"))

            @property
            def intelligent(self):
                return Capability(self._keyed_factory("supportFLIntelligent"))

            @property
            def keep_on(self):
                return Capability(self._keyed_factory("supportFLKeepOn"))

            @property
            def schedule(self):
                return Capability(self._keyed_factory("supportFLSchedule"))

            @property
            def switch(self):
                return Capability(self._keyed_factory("supportFLswitch"))

        @property
        def flood_light(self):
            return type(self).FloodLight(self._factory)

        @property
        def gop(self):
            return Capability(self._keyed_factory("supportGop"))

        @property
        def motion_detection(self):
            return Capability(self._keyed_factory("supportMd"))

        @property
        def ptz_check(self):
            return Capability(self._keyed_factory("supportPtzCheck"))

        @property
        def threshold_adjust(self):
            return Capability(self._keyed_factory("supportThresholdAdjust"))

        @property
        def white_dark(self):
            return Capability(self._keyed_factory("supportWhiteDark"))

    @property
    def supports(self):
        return type(self).Supports(self._factory)

    @property
    def video_clip(self):
        return Capability(self._keyed_factory("videoClip"))

    @property
    def watermark(self):
        return Capability(self._keyed_factory("waterMark"))

    @property
    def white_balance(self):
        return Capability(self._keyed_factory("white_balance"))

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._factory())}>"


class _Channels(Mapping[int, ChannelCapabilities]):
    """Channels"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], list]) -> None:
        super().__init__()
        self._factory = factory

    def __getitem__(self, __k: int):
        def _get():
            if (_list := self._factory()) is None:
                return None
            return _list[__k]

        return ChannelCapabilities(_get)

    def __iter__(self):
        if (_list := self._factory()) is None:
            return
        for i in range(0, len(_list)):
            yield i

    def __len__(self):
        if (_list := self._factory()) is None:
            return 0
        return len(_list)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._factory())}>"


class Capabilities(capabilities.Capabilities):
    """Capabilities"""

    __slots__ = ("_value",)

    def __init__(self, value: dict) -> None:
        super().__init__()
        self._value = value

    def _factory(self):
        return self._value

    def _keyed_factory(self, key: str):
        def _factory():
            return self._value.get(key, None)

        return _factory

    @property
    def three_g(self):
        return Capability(self._keyed_factory("3g"))

    @property
    def channels(self):
        """channels"""
        return _Channels(self._keyed_factory("abilityChn"))

    class Alarm(capabilities.Capabilities.Alarm):
        """Alarm"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        def _keyed_factory(self, key: str):
            def _factory():
                if (value := self._factory()) is None:
                    return None
                return value.get(key, None)

            return _factory

        @property
        def audio(self):
            return Capability(self._keyed_factory("alarmAudio"))

        @property
        def disconnect(self):
            return Capability(self._keyed_factory("alarmDisconnet"))

        class HDD(capabilities.Capabilities.Alarm.HDD):
            """HDDD"""

            __slots__ = ("_factory",)

            def __init__(self, factory: Callable[[], dict]) -> None:
                super().__init__()
                self._factory = factory

            def _keyed_factory(self, key: str):
                def _factory():
                    if (value := self._factory()) is None:
                        return None
                    return value.get(key, None)

                return _factory

            @property
            def error(self):
                return Capability(self._keyed_factory("alarmHddErr"))

            @property
            def full(self):
                return Capability(self._keyed_factory("alarmHddFull"))

        @property
        def hdd(self):
            return type(self).HDD(self._factory)

        @property
        def ip_conflict(self):
            return Capability(self._keyed_factory("alarmIpConflict"))

    @property
    def alarm(self):
        return type(self).Alarm(self._factory)

    @property
    def auth(self):
        return Capability(self._keyed_factory("auth"))

    @property
    def auto_maintenance(self):
        return Capability(self._keyed_factory("autoMaint"))

    @property
    def cloud_storage(self):
        return Capability(self._keyed_factory("cloudStorage"))

    @property
    def custom_audio(self):
        return Capability(self._keyed_factory("customAudio"))

    @property
    def date_format(self):
        return Capability(self._keyed_factory("dateFormat"))

    @property
    def ddns(self):
        return Capability(self._keyed_factory("ddns"))

    class Device(capabilities.Capabilities.Device):
        """Device"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        def _keyed_factory(self, key: str):
            def _factory():
                if (value := self._factory()) is None:
                    return None
                return value.get(key, None)

            return _factory

        @property
        def info(self):
            return Capability(self._keyed_factory("devInfo"))

        @property
        def name(self):
            return Capability(self._keyed_factory("devName"))

    @property
    def device(self):
        return type(self).Device(self._factory)

    @property
    def disable_autofocus(self):
        return Capability(self._keyed_factory("disableAutoFocus"))

    @property
    def disk(self):
        return Capability(self._keyed_factory("disk"))

    @property
    def display(self):
        return Capability(self._keyed_factory("display"))

    class Email(capabilities.Capabilities.Email):
        """Email"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        def _keyed_factory(self, key: str):
            def _factory():
                if (value := self._factory()) is None:
                    return None
                return value.get(key, None)

            return _factory

        @property
        def _value(self):
            return Capability(self._keyed_factory("email"))

        @property
        def value(self):
            return self._value.value

        @property
        def permissions(self):
            return self._value.permissions

        def __bool__(self):
            return bool(self.value)

        @property
        def interval(self):
            return Capability(self._keyed_factory("emailInterval"))

        @property
        def schedule(self):
            return Capability(self._keyed_factory("emailSchedule"))

    @property
    def email(self):
        return type(self).Email(self._factory)

    @property
    def config_import(self):
        return Capability(self._keyed_factory("importCfg"))

    @property
    def config_export(self):
        return Capability(self._keyed_factory("exportCfg"))

    class FTP(capabilities.Capabilities.FTP):
        """FTP"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        def _keyed_factory(self, key: str):
            def _factory():
                if (value := self._factory()) is None:
                    return None
                return value.get(key, None)

            return _factory

        @property
        def auto_dir(self):
            return Capability(self._keyed_factory("ftpAutoDir"))

        class Stream(capabilities.Capabilities.FTP.Stream):
            """Stream"""

            __slots__ = ("_factory",)

            def __init__(self, factory: Callable[[], dict]) -> None:
                super().__init__()
                self._factory = factory

            def _keyed_factory(self, key: str):
                def _factory():
                    if (value := self._factory()) is None:
                        return None
                    return value.get(key, None)

                return _factory

            @property
            def ext(self):
                return Capability(self._keyed_factory("ftpExtStream"))

            @property
            def sub(self):
                return Capability(self._keyed_factory("ftpSubStream"))

        @property
        def stream(self):
            return type(self).Stream(self._factory)

        @property
        def picture(self):
            return Capability(self._keyed_factory("ftpPic"))

        @property
        def test(self):
            return Capability(self._keyed_factory("ftpTest"))

    @property
    def ftp(self):
        return type(self).FTP(self._factory)

    @property
    def hour_format(self):
        """change hour format supported"""

        return Capability(self._keyed_factory("hourFmt"))

    @property
    def http(self):
        return Capability(self._keyed_factory("http"))

    @property
    def http_flv(self):
        return Capability(self._keyed_factory("httpFlv"))

    @property
    def https(self):
        return Capability(self._keyed_factory("https"))

    @property
    def ipc_manager(self):
        return Capability(self._keyed_factory("ipcManager"))

    @property
    def led_control(self):
        return Capability(self._keyed_factory("ledControl"))

    @property
    def local_link(self):
        return Capability(self._keyed_factory("localLink"))

    @property
    def log(self):
        return Capability(self._keyed_factory("log"))

    @property
    def media_port(self):
        return Capability(self._keyed_factory("mediaPort"))

    @property
    def ntp(self):
        return Capability(self._keyed_factory("ntp"))

    @property
    def online(self):
        return Capability(self._keyed_factory("online"))

    @property
    def onvif(self):
        return Capability(self._keyed_factory("onvif"))

    @property
    def p2p(self):
        return Capability(self._keyed_factory("p2p"))

    @property
    def performance(self):
        return Capability(self._keyed_factory("performance"))

    @property
    def pppoe(self):
        return Capability(self._keyed_factory("pppoe"))

    @property
    def push(self):
        return Capability(self._keyed_factory("push"))

    @property
    def push_schedule(self):
        return Capability(self._keyed_factory("pushSchedule"))

    @property
    def reboot(self):
        return Capability(self._keyed_factory("reboot"))

    class Record(capabilities.Capabilities.Record):
        """Record"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        def _keyed_factory(self, key: str):
            def _factory():
                if (value := self._factory()) is None:
                    return None
                return value.get(key, None)

            return _factory

        @property
        def extension_time_list(self):
            return Capability(self._keyed_factory("recExtensionTimeList"))

        @property
        def overwrite(self):
            return Capability(self._keyed_factory("recOverWrite"))

        @property
        def pack_duration(self):
            return Capability(self._keyed_factory("recPackDuration"))

        @property
        def pre_record(self):
            return Capability(self._keyed_factory("recPreRecord"))

    @property
    def record(self):
        return type(self).Record(self._factory)

    @property
    def restore(self):
        return Capability(self._keyed_factory("restore"))

    @property
    def rtmp(self):
        return Capability(self._keyed_factory("rtmp"))

    @property
    def rtsp(self):
        return Capability(self._keyed_factory("rtsp"))

    @property
    def schedule_version(self):
        return Capability(self._keyed_factory("scheduleVersion"))

    @property
    def sd_card(self):
        return Capability(self._keyed_factory("sdCard"))

    @property
    def show_qr_code(self):
        return Capability(self._keyed_factory("showQrCode"))

    @property
    def sim_module(self):
        return Capability(self._keyed_factory("simModule"))

    class Supports(capabilities.Capabilities.Supports):
        """Supports"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        class Audio(capabilities.Capabilities.Supports.Audio):
            """Audio"""

            __slots__ = ("_factory",)

            def __init__(self, factory: Callable[[], dict]) -> None:
                super().__init__()
                self._factory = factory

            class Alarm(capabilities.Capabilities.Supports.Audio.Alarm):
                """Alarm"""

                __slots__ = ("_factory",)

                def __init__(self, factory: Callable[[], dict]) -> None:
                    super().__init__()
                    self._factory = factory

                def _keyed_factory(self, key: str):
                    def _factory():
                        if (value := self._factory()) is None:
                            return None
                        return value.get(key, None)

                    return _factory

                @property
                def _value(self):
                    return Capability(self._keyed_factory("supportAudioAlarm"))

                @property
                def value(self):
                    return self._value.value

                @property
                def permissions(self):
                    return self._value.permissions

                def __bool__(self):
                    return bool(self.value)

                @property
                def enable(self):
                    return Capability(self._keyed_factory("supportAudioAlarmEnable"))

                @property
                def schedule(self):
                    return Capability(self._keyed_factory("supportAudioAlarmSchedule"))

                @property
                def task_enable(self):
                    return Capability(
                        self._keyed_factory("supportAudioAlarmTaskEnable")
                    )

            @property
            def alarm(self):
                return type(self).Alarm(self._factory)

        @property
        def audio(self):
            return type(self).Audio(self._factory)

        class Buzzer(capabilities.Capabilities.Supports.Buzzer):
            """Buzzer"""

            __slots__ = ("_factory",)

            def __init__(self, factory: Callable[[], dict]) -> None:
                super().__init__()
                self._factory = factory

            class Task(capabilities.Capabilities.Supports.Buzzer.Task):
                """Task"""

                __slots__ = ("_factory",)

                def __init__(self, factory: Callable[[], dict]) -> None:
                    super().__init__()
                    self._factory = factory

                def _keyed_factory(self, key: str):
                    def _factory():
                        if (value := self._factory()) is None:
                            return None
                        return value.get(key, None)

                    return _factory

                @property
                def _value(self):
                    return Capability(self._keyed_factory("supportBuzzerTask"))

                @property
                def value(self):
                    return self._value.value

                @property
                def permissions(self):
                    return self._value.permissions

                def __bool__(self):
                    return bool(self.value)

                @property
                def enable(self):
                    return Capability(self._keyed_factory("supportBuzzerEnable"))

            def _keyed_factory(self, key: str):
                def _factory():
                    if (value := self._factory()) is None:
                        return None
                    return value.get(key, None)

                return _factory

            @property
            def _value(self):
                return Capability(self._keyed_factory("supportBuzzer"))

            @property
            def task(self):
                return type(self).Task(self._factory)

            @property
            def value(self):
                """value"""
                return self._value.value

            @property
            def permissions(self):
                """permissions"""
                return self._value.permissions

        @property
        def buzzer(self):
            return type(self).Buzzer(self._factory)

        class Email(capabilities.Capabilities.Supports.Email):
            """Email"""

            __slots__ = ("_factory",)

            def __init__(self, factory: Callable[[], dict]) -> None:
                super().__init__()
                self._factory = factory

            def _keyed_factory(self, key: str):
                def _factory():
                    if (value := self._factory()) is None:
                        return None
                    return value.get(key, None)

                return _factory

            @property
            def enable(self):
                return Capability(self._keyed_factory("supportEmailEnable"))

            @property
            def task_enable(self):
                return Capability(self._keyed_factory("supportEmailTaskEnable"))

        @property
        def email(self):
            return type(self).Email(self._factory)

        class FTP(capabilities.Capabilities.Supports.FTP):
            """FTP"""

            __slots__ = ("_factory",)

            def __init__(self, factory: Callable[[], dict]) -> None:
                super().__init__()
                self._factory = factory

            class Cover(capabilities.Capabilities.Supports.FTP.Cover):
                """Cover"""

                __slots__ = ("_factory",)

                def __init__(self, factory: Callable[[], dict]) -> None:
                    super().__init__()
                    self._factory = factory

                def _keyed_factory(self, key: str):
                    def _factory():
                        if (value := self._factory()) is None:
                            return None
                        return value.get(key, None)

                    return _factory

                @property
                def picture(self):
                    return Capability(self._keyed_factory("supportFtpCoverPicture"))

                @property
                def video(self):
                    return Capability(self._keyed_factory("supportFtpCoverVideo"))

            def _keyed_factory(self, key: str):
                def _factory():
                    if (value := self._factory()) is None:
                        return None
                    return value.get(key, None)

                return _factory

            @property
            def cover(self):
                return type(self).Cover(self._factory)

            @property
            def dir_YM(self):
                return Capability(self._keyed_factory("supportFtpDirYM"))

            @property
            def enable(self):
                return Capability(self._keyed_factory("supportFtpEnable"))

            class Picture(capabilities.Capabilities.Supports.FTP.Picture):
                """Picture"""

                __slots__ = ("_factory",)

                def __init__(self, factory: Callable[[], dict]) -> None:
                    super().__init__()
                    self._factory = factory

                def _keyed_factory(self, key: str):
                    def _factory():
                        if (value := self._factory()) is None:
                            return None
                        return value.get(key, None)

                    return _factory

                @property
                def capture_mode(self):
                    return Capability(self._keyed_factory("supportFtpPicCaptureMode"))

                @property
                def custom_resolution(self):
                    return Capability(self._keyed_factory("supportFtpPicResoCustom"))

                @property
                def swap(self):
                    return Capability(self._keyed_factory("supportFtpPictureSwap"))

            @property
            def picture(self):
                return type(self).Picture(self._factory)

            class Task(capabilities.Capabilities.Supports.FTP.Task):
                """Task"""

                __slots__ = ("_factory",)

                def __init__(self, factory: Callable[[], dict]) -> None:
                    super().__init__()
                    self._factory = factory

                def _keyed_factory(self, key: str):
                    def _factory():
                        if (value := self._factory()) is None:
                            return None
                        return value.get(key, None)

                    return _factory

                @property
                def _value(self):
                    return Capability(self._keyed_factory("supportFtpTask"))

                @property
                def value(self):
                    return self._value.value

                @property
                def permissions(self):
                    return self._value.permissions

                def __bool__(self):
                    return bool(self.value)

                @property
                def enable(self):
                    return Capability(self._keyed_factory("supportFtpTaskEnable"))

            def _keyed_factory(self, key: str):
                def _factory():
                    if (value := self._factory()) is None:
                        return None
                    return value.get(key, None)

                return _factory

            @property
            def task(self):
                return type(self).Task(self._factory)

            @property
            def video_swap(self):
                return Capability(self._keyed_factory("supportFtpVideoSwap"))

            @property
            def ftps_encrypt(self):
                return Capability(self._keyed_factory("supportFtpsEncrypt"))

        def _keyed_factory(self, key: str):
            def _factory():
                if (value := self._factory()) is None:
                    return None
                return value.get(key, None)

            return _factory

        @property
        def ftp(self):
            return type(self).FTP(self._factory)

        @property
        def http_enable(self):
            return Capability(self._keyed_factory("supportHttpEnable"))

        @property
        def https_enable(self):
            return Capability(self._keyed_factory("supportHttpsEnable"))

        @property
        def onvif_enable(self):
            return Capability(self._keyed_factory("supportOnvifEnable"))

        @property
        def push_interval(self):
            return Capability(self._keyed_factory("supportPushInterval"))

        class Record(capabilities.Capabilities.Supports.Record):
            """Record"""

            __slots__ = ("_factory",)

            def __init__(self, factory: Callable[[], dict]) -> None:
                super().__init__()
                self._factory = factory

            def _keyed_factory(self, key: str):
                def _factory():
                    if (value := self._factory()) is None:
                        return None
                    return value.get(key, None)

                return _factory

            @property
            def schedule_enable(self):
                return Capability(self._keyed_factory("supportRecScheduleEnable"))

            @property
            def enable(self):
                return Capability(self._keyed_factory("supportRecordEnable"))

        @property
        def record(self):
            return type(self).Record(self._factory)

        @property
        def rtmp_enable(self):
            return Capability(self._keyed_factory("supportRtmpEnable"))

        @property
        def rtsp_enable(self):
            return Capability(self._keyed_factory("supportRtspEnable"))

    @property
    def supports(self):
        return type(self).Supports(self._factory)

    @property
    def talk(self):
        return Capability(self._keyed_factory("talk"))

    @property
    def time(self):
        return Capability(self._keyed_factory("time"))

    @property
    def tv_system(self):
        return Capability(self._keyed_factory("tvSystem"))

    @property
    def upgrade(self):
        return Capability(self._keyed_factory("upgrade"))

    @property
    def upnp(self):
        return Capability(self._keyed_factory("upnp"))

    @property
    def user(self):
        return Capability(self._keyed_factory("user"))

    @property
    def video_clip(self):
        return Capability(self._keyed_factory("videoClip"))

    class Wifi(capabilities.Capabilities.Wifi):
        """WIFI"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        def _keyed_factory(self, key: str):
            def _factory():
                if (value := self._factory()) is None:
                    return None
                return value.get(key, None)

            return _factory

        @property
        def _value(self):
            return Capability(self._keyed_factory("wifi"))

        @property
        def value(self):
            return self._value.value

        @property
        def permissions(self):
            return self._value.permissions

        def __bool__(self):
            return bool(self.value)

        @property
        def testable(self):
            return Capability(self._keyed_factory("wifiTest"))

    @property
    def wifi(self):
        return type(self).Wifi(self._factory)

    def update(self, value: "Capabilities"):
        if not isinstance(value, type(self)):
            raise TypeError("Can only update from another Capabilities")
        # pylint: disable=protected-access
        self._value = value._value
        return self

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._value)}>"



"""REST Network Models"""

from typing import Callable, Final, Mapping
from async_reolink.api.network import typings

from .typings import STR_LINKTYPES_MAP

# pylint: disable=missing-function-docstring


class IPInfo(typings.IPInfo):
    """REST IP Info"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        self._factory = factory

    @property
    def gateway(self) -> str:
        if (value := self._factory()) is None:
            return None
        return value.get("gateway", None)

    @property
    def address(self) -> str:
        if (value := self._factory()) is None:
            return None
        return value.get("ip", None)

    @property
    def mask(self) -> str:
        if (value := self._factory()) is None:
            return None
        return value.get("mask", None)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._factory())}>"


class LinkInfo(typings.LinkInfo):
    """REST Link Info"""

    __slots__ = ("_value",)

    def __init__(self, value: dict) -> None:
        if value is None:
            value = {}
        self._value = value

    @property
    def active_link(self) -> str:
        return self._value.get("activeLink", None)

    @property
    def mac(self) -> str:
        return self._value.get("mac", None)

    @property
    def type(self):
        if (value := self._value.get("type", None)) is None:
            return None
        return STR_LINKTYPES_MAP[value]

    @property
    def ip(self):  # pylint: disable=invalid-name
        def _get():
            return self._value.get("static", None)

        return IPInfo(_get)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._value)}>"


_CHANNEL_KEY: Final = "channel"


class ChannelStatus(typings.ChannelStatus):
    """REST Channel Status"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        self._factory = factory

    @property
    def channel_id(self) -> int:
        if (value := self._factory()) is None:
            return None
        return value.get(_CHANNEL_KEY, None)

    @property
    def name(self) -> str:
        if (value := self._factory()) is None:
            return ""
        return value.get("name", "")

    @property
    def online(self) -> bool:
        if (value := self._factory()) is None:
            return 0
        return value.get("online", 0)

    @property
    def type(self) -> str:
        if (value := self._factory()) is None:
            return None
        return value.get("typeInfo", None)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._factory())}>"


class ChannelStatuses(Mapping[int, ChannelStatus]):
    """Channel Statuses"""

    __slots__ = ("_value",)

    def __init__(self, value: dict) -> None:
        super().__init__()
        if value is None:
            value = {}
        self._value = value

    def _factory(self):
        return self._value

    @staticmethod
    def _get_channel(value: dict) -> int:
        return value.get(_CHANNEL_KEY, None) if value is not None else None

    def _get_channels(self) -> list:
        return self._value.get("status", None)

    def _channel_getter(self, channel_id: int):
        def _is_channel(value: dict):
            return type(self)._get_channel(value) == channel_id

        def _factory() -> dict:
            if (channels := self._get_channels()) is None:
                return None
            return next(filter(_is_channel, channels), None)

        return _factory

    def __getitem__(self, __k: int):
        return ChannelStatus(self._channel_getter(__k))

    def __iter__(self):
        if (channels := self._get_channels()) is None:
            return
        for _d in channels:
            if (channel := type(self)._get_channel(_d)) is not None:
                yield channel

    def __contains__(self, __o: int) -> bool:
        return self._channel_getter(__o)() is not None

    def __len__(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("count", 0)

    def update(self, value: "ChannelStatuses"):
        if not isinstance(value, type(self)):
            raise TypeError("Can only update from another ChannelStatuses")
        # pylint: disable=protected-access
        self._value = value._value
        return self

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._value)}>"


class P2PInfo(typings.P2PInfo):
    """REST P2P Info"""

    __slots__ = ("_value",)

    def __init__(self, value: dict) -> None:
        if value is None:
            value = {}
        self._value = value

    @property
    def enabled(self) -> bool:
        return self._value.get("enable", 0)

    @property
    def uid(self) -> str:
        return self._value.get("uid", None)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._value)}>"


class NetworkPort(typings.NetworkPort):
    """REST Network Port"""

    __slots__ = ("_factory", "_prefix")

    def __init__(self, prefix: str, factory: Callable[[], dict]) -> None:
        self._prefix = prefix
        self._factory = factory

    @property
    def value(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get(self._prefix + "Port", 0)

    @property
    def enabled(self) -> bool:
        if (value := self._factory()) is None:
            return 0
        # some versions dont support Enable so the port cannot be disabled, so we default to true if Port exists
        return value.get(
            self._prefix + "Enable", 1 if self._prefix + "Port" in value else 0
        )

    def __bool__(self):
        return bool(self.enabled and self.value)

    def __int__(self):
        return self.value if self.enabled else 0

    def __str__(self):
        return str(self.value) if self.enabled else ""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: enabled={bool(self.enabled)}, value={self.value}>"


class NetworkPorts(typings.NetworkPorts):
    """REST Network Ports"""

    __slots__ = ("_value",)

    def __init__(self, value: dict) -> None:
        if value is None:
            value = {}
        self._value = value

    def _factory(self):
        return self._value

    @property
    def media(self):
        return NetworkPort("media", self._factory)

    @property
    def http(self):
        return NetworkPort("http", self._factory)

    @property
    def https(self):
        return NetworkPort("https", self._factory)

    @property
    def onvif(self):
        return NetworkPort("onvif", self._factory)

    @property
    def rtsp(self):
        return NetworkPort("rtsp", self._factory)

    @property
    def rtmp(self):
        return NetworkPort("rtmp", self._factory)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {repr(self._factory())}>"



"""REST LED Mixin"""

from typing import Mapping, Sequence
from async_reolink.api import led
from async_reolink.api.led.typings import LightStates, LightingSchedule, WhiteLedInfo
from async_reolink.api.ai.typings import AITypes
from async_reolink.api.typings import PercentValue

from ..commands import led as commands


class LED(led.LED):
    """REST LED Mixin"""

    def _create_get_ir_lights_request(self, channel: int):
        return commands.GetIrLightsRequest(channel)

    def _create_set_ir_lights_request(self, state: LightStates, channel: int):
        return commands.SetIrLightsRequest(state, channel)

    def _create_get_power_led_request(self, channel: int):
        return commands.GetPowerLedRequest(channel)

    def _create_set_power_led_request(self, state: LightStates, channel: int):
        return commands.SetPowerLedRequest(state, channel)

    def _create_get_white_led_request(self, channel: int):
        return commands.GetWhiteLedRequest(channel)

    def _create_set_white_led_request(
        self,
        info: WhiteLedInfo,
        channel: int,
    ):
        req = commands.SetWhiteLedRequest(channel)
        req.info.state = info.state
        if info.brightness is not None:
            req.info.brightness = info.brightness
        if info.brightness_state is not None:
            req.info.brightness_state = info.brightness_state
        if info.lighting_schedule is not None:
            req.info.lighting_schedule = info.lighting_schedule
        if info.ai_detection_type is not None:
            req.info.ai_detection_type = info.ai_detection_type
        return req



""" ReoLink REST API """

from . import (
    ai,
    alarm,
    encoding,
    connection,
    led,
    ptz,
    network,
    system,
    record,
    security,
    video,
)

from .__version__ import __version__


class Client(
    connection.Connection,
    security.Security,
    system.System,
    network.Network,
    video.Video,
    encoding.Encoding,
    record.Record,
    alarm.Alarm,
    ai.AI,
    led.LED,
    ptz.PTZ,
):
    """Rest API Client"""

    def __init__(self, session_factory: connection.SessionFactory = None) -> None:
        super().__init__(session_factory=session_factory)


# __all__ = ["Client", "__version__"]



"""REST Alarm Mixin"""

from async_reolink.api.alarm import Alarm as BaseAlarm

from ..commands import alarm


class Alarm(BaseAlarm):
    """REST Alarm Mixin"""

    def _create_get_md_state(self, channel: int):
        return alarm.GetMotionStateRequest(channel)



"""Encoding typings"""

from types import MappingProxyType
from typing import Final

from async_reolink.api.typings import StreamTypes


STR_STREAMTYPES_MAP: Final = MappingProxyType(
    {_e.name.lower() + "Stream": _e for _e in StreamTypes}
)

STREAMTYPES_STR_MAP: Final = MappingProxyType(
    {_v: _k for _k, _v in STR_STREAMTYPES_MAP.items()}
)



"""AI Models"""

from typing import (
    Callable,
    Final,
    Mapping,
    MutableMapping,
)
from async_reolink.api.ai import typings

from ..typings import FactoryValue
from .typings import AITYPES_STR_MAP

# pylint: disable=too-few-public-methods
# pylint:disable=missing-function-docstring


class AlarmState(typings.AlarmState):
    """Alarm State"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        super().__init__()
        self._factory = factory

    @property
    def state(self) -> bool:
        if (value := self._factory()) is not None:
            return value.get("alarm_state", 0)
        return False

    @property
    def supported(self) -> bool:
        if (value := self._factory()) is not None:
            return value.get("support", 0)
        return False

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._factory())}>"


class State(Mapping[typings.AITypes, AlarmState]):
    """AI State"""

    __slots__ = ("_value",)

    def __init__(self, value: dict) -> None:
        super().__init__()
        if value is None:
            value = {}
        self._value = value

    def __getitem__(self, __k: typings.AITypes):
        def _factory() -> dict:
            return self._value.get(AITYPES_STR_MAP[__k], None)

        return AlarmState(_factory)

    def __contains__(self, __o: typings.AITypes):
        return AITYPES_STR_MAP[__o] in self._value

    def __iter__(self):
        for _k, _v in AITYPES_STR_MAP.items():
            if _v in self._value:
                yield _k

    def __len__(self):
        return len(AITYPES_STR_MAP.values() & self._value.keys())

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._value)}>"


class AITypesMap(Mapping[typings.AITypes, bool]):
    """AI Types Map"""

    __slots__ = ("_factory",)

    def __init__(self, factory: FactoryValue[dict]) -> None:
        super().__init__()
        self._factory = factory

    def __getitem__(self, __k: typings.AITypes) -> bool:
        if (_map := self._factory()) is not None:
            return _map.get(AITYPES_STR_MAP[__k], 0)
        return False

    def __iter__(self):
        if (_map := self._factory()) is None:
            return
        for _e in typings.AITypes:
            if _map is not None and AITYPES_STR_MAP[_e] in _map:
                yield _e

    def __len__(self):
        if (_map := self._factory()) is not None:
            return _map.__len__()
        return 0

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._factory())}>"


class MutableAITypesMap(AITypesMap, MutableMapping[typings.AITypes, bool]):
    """Mutable AI Types Map"""

    def __setitem__(self, __k: typings.AITypes, __v: bool) -> None:
        if (_map := self._factory(True)) is None:
            raise KeyError()
        _map[AITYPES_STR_MAP[__k]] = int(__v)

    def __delitem__(self, __v: typings.AITypes) -> None:
        if (_map := self._factory(True)) is None:
            raise KeyError()
        del _map[AITYPES_STR_MAP[__v]]

    def clear(self) -> None:
        if (_map := self._factory()) is None:
            return
        _map.clear()


_DETECT_TYPE_KEY: Final = "AiDetectType"
_TYPE_KEY: Final = "type"
_AI_TRACK_KEY: Final = "aiTrack"
_TRACK_TYPE_KEY: Final = "trackType"


class Config(typings.Config):
    """AI Configuration"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        super().__init__()
        if not callable(factory):
            value = factory
            if value is None:
                value = {}

            def _factory():
                return value

            factory = _factory
        self._factory = factory

    def _get_detect_type(self) -> dict:
        return (
            value.get(_DETECT_TYPE_KEY, None)
            if (value := self._factory()) is not None
            else None
        )

    def _get_type(self) -> dict:
        return (
            value.get(_TYPE_KEY, None)
            if (value := self._factory()) is not None
            else None
        )

    @property
    def detect_type(self):
        """detect type"""

        def _get():
            if (_dict := self._get_detect_type()) is not None:
                return _dict
            if (_dict := self._get_type()) is not None:
                return _dict
            return None

        return AITypesMap(_get)

    @property
    def ai_track(self) -> bool:
        return (
            value.get(_AI_TRACK_KEY, 0) if (value := self._factory()) is not None else 0
        )

    def _get_track_type(self) -> dict:
        return (
            value.get(_TRACK_TYPE_KEY, None)
            if (value := self._factory()) is not None
            else None
        )

    @property
    def track_type(self):
        return AITypesMap(self._get_track_type)

    def update(self, value: "Config"):
        if not isinstance(value, type(self)):
            raise TypeError("Can only update from another Config")
        # pylint: disable=protected-access
        _value = value._factory()

        def _factory():
            return _value

        self._factory = _factory
        return self

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._factory())}>"


class MutableConfig(Config):
    """Mutable AI Configuration"""

    def __init__(self, factory: FactoryValue[dict]) -> None:
        super().__init__(factory)
        self._factory = factory

    def _get_dict(self, key: str, create=False) -> dict:
        if (parameter := self._factory(create)) is None:
            return None
        if create and key not in parameter:
            return parameter.setdefault(key, {})
        return parameter.get(key, None)

    @property
    def _value(self):
        return self._factory(True)

    @property
    def _detect_type(self):
        return self._get_dict(_DETECT_TYPE_KEY, True)

    @property
    def _type(self):
        return self._get_dict(_TYPE_KEY, True)

    @_type.setter
    def _type(self, value):
        self._value[_TYPE_KEY] = value

    @property
    def detect_type(self):
        """detect type"""

        def _get(ensure: bool):
            if (value := self._get_dict(_DETECT_TYPE_KEY, ensure)) is None:
                if (value := self._get_dict(_TYPE_KEY, ensure)) is None:
                    return None
            return value

        return MutableAITypesMap(_get)

    def _update_map(self, _map: MutableAITypesMap, value):
        _map.clear()
        if isinstance(value, Mapping):
            _map.update(**value)
        if isinstance(value, typings.AITypes):
            _map[value] = True
        for item in value:
            _map[typings.AITypes(item)] = True

    @detect_type.setter
    def detect_type(self, value):
        self._update_map(self.detect_type, value)

    @property
    def _has_track_type(self):
        if (parameter := self._factory()) is None:
            return False
        return _TRACK_TYPE_KEY in parameter

    @property
    def track_type(self):
        """track type"""

        def _get(ensure: bool):
            if self._has_track_type or ensure:
                return self._value.get(_TRACK_TYPE_KEY, {})
            return None

        return MutableAITypesMap(_get)

    @track_type.setter
    def track_type(self, value):
        self._update_map(self.track_type, value)



"""System models"""

from typing import Callable, TypeVar
from async_reolink.api.commands import system
from async_reolink.api.system import typings

# pylint: disable=missing-function-docstring
# pylint: disable=too-few-public-methods

_T = TypeVar("_T")


class DeviceInfo(typings.DeviceInfo):
    """Device Info"""

    __slots__ = ("_value",)

    def __init__(self, value: dict) -> None:
        super().__init__()
        if value is None:
            value = {}
        self._value = value

    def _factory(self):
        return self._value

    class IO(typings.DeviceInfo.IO):
        """IO"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        @property
        def inputs(self):
            if (value := self._factory()) is None:
                return 0
            return value.get("IOInputNum", 0)

        @property
        def outputs(self):
            if (value := self._factory()) is None:
                return 0
            return value.get("IOOutputNum", 0)

    @property
    def io(self):
        return type(self).IO(self._factory)

    class Version(typings.DeviceInfo.Version):
        """Versions"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        @property
        def firmware(self):
            if (value := self._factory()) is None:
                return ""
            return value.get("firmVer", "")

        @property
        def framework(self):
            if (value := self._factory()) is None:
                return ""
            return value.get("frameworkVer", "")

        @property
        def hardware(self):
            if (value := self._factory()) is None:
                return ""
            return value.get("hardVer", "")

        @property
        def config(self):
            if (value := self._factory()) is None:
                return ""
            return value.get("cfgVer", "")

    @property
    def version(self):
        return type(self).Version(self._factory)

    @property
    def audio_sources(self):
        if (value := self._factory()) is None:
            return 0
        return value.get("audioNum", 0)

    @property
    def build_day(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("buildDay", "")

    @property
    def channels(self):
        if (value := self._factory()) is None:
            return 0
        return value.get("channelNum", 0)

    @property
    def detail(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("detail", "")

    @property
    def disks(self):
        if (value := self._factory()) is None:
            return 0
        return value.get("diskNum", 0)

    @property
    def model(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("model", "")

    @property
    def name(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("name", "")

    @property
    def type(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("type", "")

    @property
    def wifi(self):
        if (value := self._factory()) is None:
            return False
        return value.get("wifi", False)

    # @property
    def B845(self):
        if (value := self._factory()) is None:
            return 0
        return value.get("", 0)

    @property
    def exact_type(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("exactType", "")

    @property
    def serial(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("serial", "")

    @property
    def pak_suffix(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("pakSuffix", "")

    def update(self, value: "DeviceInfo"):
        if not isinstance(value, type(self)):
            raise TypeError("Can only update from another DeviceInfo")
        # pylint: disable=protected-access
        self._value = value._value
        return self


class DaylightSavingsTimeInfo(system.DaylightSavingsTimeInfo):
    """Dalight Savings Time Info"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        super().__init__()
        self._factory = factory

    @property
    def enabled(self):
        if (value := self._factory()) is None:
            return False
        return value.get("enable", False)

    @property
    def hour_offset(self):
        if (value := self._factory()) is None:
            return 0
        return value.get("offset", 0)

    class StartDateTime(system.DaylightSavingsTimeInfo.TimeInfo):
        """Start Time"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        @property
        def month(self):
            if (value := self._factory()) is None:
                return 0
            return value.get("startMon", 0)

        @property
        def hour(self):
            if (value := self._factory()) is None:
                return 0
            return value.get("startHour", 0)

        @property
        def minute(self):
            if (value := self._factory()) is None:
                return 0
            return value.get("startMin", 0)

        @property
        def second(self):
            if (value := self._factory()) is None:
                return 0
            return value.get("startSec", 0)

        @property
        def week(self):
            if (value := self._factory()) is None:
                return 0
            return value.get("startWeek", 0)

        @property
        def weekday(self):
            if (value := self._factory()) is None:
                return 0
            return value.get("startWeekday", 0)

    @property
    def start(self):
        return type(self).StartDateTime(self._factory)

    class EndDateTime(system.DaylightSavingsTimeInfo.TimeInfo):
        """End Time"""

        __slots__ = ("_factory",)

        def __init__(self, factory: Callable[[], dict]) -> None:
            super().__init__()
            self._factory = factory

        @property
        def month(self):
            if (value := self._factory()) is None:
                return 0
            return value.get("endMon", 0)

        @property
        def hour(self):
            if (value := self._factory()) is None:
                return 0
            return value.get("endHour", 0)

        @property
        def minute(self):
            if (value := self._factory()) is None:
                return 0
            return value.get("endMin", 0)

        @property
        def second(self):
            if (value := self._factory()) is None:
                return 0
            return value.get("endSec", 0)

        @property
        def week(self):
            if (value := self._factory()) is None:
                return 0
            return value.get("endWeek", 0)

        @property
        def weekday(self):
            if (value := self._factory()) is None:
                return 0
            return value.get("endWeekday", 0)

    @property
    def end(self):
        return type(self).EndDateTime(self._factory)


class TimeInfo(system.TimeInfo):
    """Device Time"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        super().__init__()
        self._factory = factory

    @property
    def year(self):
        if (value := self._factory()) is None:
            return 0
        return value.get("year", 0)

    @property
    def month(self):
        if (value := self._factory()) is None:
            return 0
        return value.get("mon", 0)

    @property
    def day(self):
        if (value := self._factory()) is None:
            return 0
        return value.get("day", 0)

    @property
    def hour(self):
        if (value := self._factory()) is None:
            return 0
        return value.get("hour", 0)

    @property
    def minute(self):
        if (value := self._factory()) is None:
            return 0
        return value.get("min", 0)

    @property
    def second(self):
        if (value := self._factory()) is None:
            return 0
        return value.get("sec", 0)

    @property
    def hour_format(self):
        if (value := self._factory()) is None:
            return typings.HourFormat.HR_12
        return value.get("hourFmt", typings.HourFormat.HR_12)

    @property
    def date_format(self):
        if (value := self._factory()) is None:
            return "DD/MM/YYYY"
        return value.get("timeFmt", "DD/MM/YYYY")

    @property
    def timezone_offset(self):
        if (value := self._factory()) is None:
            return 0
        return value.get("timeZone", 0)



"""REST PTZ Commands"""

from typing import (
    TYPE_CHECKING,
    Callable,
    Final,
    Generator,
    Iterable,
    Mapping,
    MutableSequence,
    TypeGuard,
    TypeVar,
    cast,
)
from async_reolink.api.commands import ptz
from async_reolink.api.ptz import typings

from ..ptz.typings import (
    OPERATION_STR_MAP,
    STR_OPERATION_MAP,
    STR_ZOOMOPERATION_MAP,
    ZOOMOPERATION_STR_MAP,
)

from ..ptz.models import (
    MutablePreset,
    Preset,
    _ID_KEY,
    MutablePatrol,
    Patrol,
    MutableTrack,
    Track,
    ZoomFocus,
)

from ..typings import FactoryValue

from . import (
    _CHANNEL_KEY,
    CommandRequest,
    CommandRequestWithChannel,
    CommandResponse,
    CommandResponseTypes,
)

# pylint:disable=missing-function-docstring

_T = TypeVar("_T")


class _DictList(Mapping[int, _T]):
    __slots__ = ("_get_value", "_factory")

    def __init__(
        self, gettter: Callable[[], list], factory: Callable[[Callable[[], dict]], _T]
    ) -> None:
        self._get_value = gettter
        self._factory = factory

    def _get_item(self, __k: int) -> dict:
        return (
            next(
                (
                    _d
                    for _d in value
                    if isinstance(_d, dict) and _d.get(_ID_KEY, None) == __k
                ),
                None,
            )
            if (value := self._get_value()) is not None
            else None
        )

    def __getitem__(self, __k: int):
        def _factory():
            return self._get_item(__k)

        return self._factory(_factory)

    def __iter__(self) -> Generator[int, None, None]:
        if (value := self._get_value()) is None:
            return
        for _d in value:
            if isinstance(_d, dict) and (__k := _d.get(_ID_KEY, None)) is not None:
                yield __k

    def __contains__(self, __o: int):
        return self._get_item(__o) is not None

    def __len__(self) -> int:
        if (value := self._get_value()) is None:
            return 0
        return len(value)


class GetPresetRequest(CommandRequestWithChannel, ptz.GetPresetRequest):
    """REST Get Presets Request"""

    __slots__ = ()

    COMMAND: Final = "GetPtzPreset"

    def __init__(
        self,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id


_PRESET_KEY: Final = "PtzPreset"


class GetPresetResponse(CommandResponse, ptz.GetPresetResponse, test="is_response"):
    """Get Presets Response"""

    __slots__ = ()

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetPresetRequest.COMMAND)

    def _get_sub_value(self) -> list:
        return (
            value.get(_PRESET_KEY, None)
            if (value := self._get_value()) is not None
            else None
        )

    @property
    def channel_id(self) -> int:
        if (_list := self._get_sub_value()) is None:
            return None
        if (value := next(_list, None)) is None:
            return None
        if TYPE_CHECKING:
            value = cast(dict, value)
        return value.get(_CHANNEL_KEY, None)

    @property
    def presets(self):
        return _DictList(self._get_sub_value, Preset)


class SetPresetRequest(CommandRequest, ptz.SetPresetRequest):
    """Set Preset Request"""

    __slots__ = ()

    COMMAND: Final = "SetPtzPreset"

    def __init__(
        self,
        preset: typings.Preset = None,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        if preset is not None:
            self.preset = preset

    def _get_sub_value(self, create=False) -> dict:
        _key: Final = _PRESET_KEY
        if (parameter := self._get_parameter(create)) is None:
            return None
        if _key in parameter or not create:
            return parameter.get(_key, None)
        return parameter.setdefault(_key, {})

    @property
    def preset(self):
        return MutablePreset(self._get_sub_value)

    @preset.setter
    def preset(self, value):
        if not isinstance(value, MutablePreset):
            value = MutablePreset(value)
        # pylint: disable=protected-access
        self._parameter[_PRESET_KEY] = value._factory(True)


_PATROL_KEY: Final = "PtzPatrol"


class GetPatrolRequest(CommandRequestWithChannel, ptz.GetPatrolRequest):
    """Get Patrol"""

    __slots__ = ()

    COMMAND: Final = "GetPtzPatrol"

    def __init__(
        self,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id


class GetPatrolResponse(CommandResponse, ptz.GetPatrolResponse, test="is_response"):
    """Get Patrol Response"""

    __slots__ = ()

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetPatrolRequest.COMMAND)

    def _get_sub_value(self) -> list:
        return (
            value.get(_PATROL_KEY, None)
            if (value := self._get_value()) is not None
            else None
        )

    @property
    def channel_id(self) -> int:
        if (_list := self._get_sub_value()) is None:
            return None
        if (value := next(_list, None)) is None:
            return None
        if TYPE_CHECKING:
            value = cast(dict, value)
        return value.get(_CHANNEL_KEY, None)

    @property
    def patrols(self):
        return _DictList(self._get_sub_value, Patrol)


class SetPatrolRequest(CommandRequest, ptz.SetPatrolRequest):
    """Set  Patrol"""

    __slots__ = ()

    COMMAND: Final = "SetPtzPatrol"

    def __init__(
        self,
        patrol: typings.Patrol = None,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        if patrol is not None:
            self.patrol = patrol

    def _get_sub_value(self, create=False) -> dict:
        _key: Final = _PATROL_KEY
        if (parameter := self._get_parameter(create)) is None:
            return None
        if _key in parameter or not create:
            return parameter.get(_key, None)
        return parameter.setdefault(_key, {})

    @property
    def patrol(self):
        return MutablePatrol(self._get_sub_value)

    @patrol.setter
    def patrol(self, value):
        if not isinstance(value, MutablePatrol):
            value = MutablePatrol(value)
        # pylint: disable=protected-access
        self._parameter[_PATROL_KEY] = value._factory(True)


_TATTERN_KEY: Final = "PtzTattern"


class GetTatternRequest(CommandRequestWithChannel, ptz.GetTatternRequest):
    """Get Tattern"""

    __slots__ = ()

    COMMAND: Final = "GetPtzTattern"

    def __init__(
        self,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id


class GetTatternResponse(CommandResponse, ptz.GetTatternResponse):
    """Get Tattern Response"""

    __slots__ = ()

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetTatternRequest.COMMAND)

    def _get_sub_value(self) -> list:
        return (
            value.get(_TATTERN_KEY, None)
            if (value := self._get_value()) is not None
            else None
        )

    @property
    def channel_id(self) -> int:
        return (
            value.get(_CHANNEL_KEY, 0)
            if (value := self._get_sub_value()) is not None
            else 0
        )

    @property
    def tracks(self):
        return _DictList(self._get_sub_value, Track)


class _MutableTracks(MutableSequence[MutableTrack]):
    __slots__ = ("_factory",)

    def __init__(self, factory: FactoryValue[list]) -> None:
        super().__init__()
        self._factory = factory

    def __getitem__(self, __k: int):
        def _factory():
            if (value := self._factory()) is None:
                return None
            return value[__k]

        return MutableTrack(_factory)

    def __setitem__(self, __k: int, __v: MutableTrack):
        if (value := self._factory(True)) is not None:
            value[__k] = __v._factory(True)

    def __delitem__(self, __k: int):
        if (value := self._factory()) is not None:
            del value[__k]

    def _insert(self, index: int, value: dict):
        if (_value := self._factory(True)) is not None:
            _value.insert(index, value)

    def append(self, value: typings.Track) -> None:
        return super().append(value)

    def insert(self, index: int, value: typings.Track):
        if not isinstance(value, MutableTrack):
            value = MutableTrack(value)
        # pylint: disable=protected-access
        self._insert(index, value._factory(True))

    def clear(self) -> None:
        if (value := self._factory()) is not None:
            value.clear()

    def __len__(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return len(value)


class SetTatternRequest(CommandRequest, ptz.SetTatternRequest):
    """Set PTZ Tattern"""

    __slots__ = ()

    COMMAND: Final = "SetPtzTattern"

    def __init__(
        self,
        *tracks: typings.Track,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id
        if len(tracks) > 0:
            _tracks = self.tracks
            for track in tracks:
                _tracks.append(track)

    def _get_sub_value(self, create=False) -> dict:
        _key: Final = _TATTERN_KEY
        if (parameter := self._get_parameter(create)) is None:
            return None
        if _key in parameter or not create:
            return parameter.get(_key, None)
        return parameter.setdefault(_key, {})

    @property
    def _tattern(self):
        return self._get_sub_value(True)

    @GetTatternRequest.channel_id.setter
    def channel_id(self, value):
        self._tattern[_CHANNEL_KEY] = value

    @property
    def tracks(self):
        def _factory(create=False):
            _key: Final = "track"
            if (value := self._get_sub_value(create)) is None:
                return None
            if _key in value or not create:
                return value.get(_key, None)
            return value.setdefault(_key, [])

        return _MutableTracks(_factory)

    @tracks.setter
    def tracks(self, value: Iterable[typings.Track]):
        _tracks = self.tracks
        _tracks.clear()
        for track in value:
            _tracks.append(track)


class GetAutoFocusRequest(CommandRequestWithChannel, ptz.GetAutoFocusRequest):
    """Get PTZ AutoFocus"""

    __slots__ = ()

    COMMAND: Final = "GetAutoFocus"

    def __init__(
        self,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id


_AUTO_FOCUS_KEY: Final = "AutoFocus"
_DISABLE_KEY: Final = "disable"


class GetAutoFocusResponse(
    CommandResponse, ptz.GetAutoFocusResponse, test="is_response"
):
    """Get PTZ AutoFocus Response"""

    __slots__ = ()

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetAutoFocusRequest.COMMAND)

    def _get_sub_value(self) -> dict:
        return (
            value.get(_AUTO_FOCUS_KEY, None)
            if (value := self._get_value()) is not None
            else None
        )

    channel_id = GetTatternResponse.channel_id

    @property
    def disabled(self) -> bool:
        if (value := self._get_sub_value()) is None:
            return 0
        return value.get(_DISABLE_KEY, 0)


class SetAutoFocusRequest(CommandRequestWithChannel, ptz.SetAutoFocusRequest):
    """Set PTZ AutoFocus"""

    __slots__ = ()

    COMMAND: Final = "AutoFocus"

    def __init__(
        self,
        disabled: bool,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id
        self.disabled = disabled

    def _get_sub_value(self, create=False) -> dict:
        _key: Final = _AUTO_FOCUS_KEY
        if (parameter := self._get_parameter(create)) is None:
            return None
        if _key in parameter or not create:
            return parameter.get(_key, None)
        return parameter.setdefault(_key, {})

    @property
    def _sub_value(self):
        return self._get_sub_value(True)

    channel_id = SetTatternRequest.channel_id

    @GetAutoFocusResponse.disabled.setter
    def disabled(self, value):
        self._sub_value[_DISABLE_KEY] = int(value)


class GetZoomFocusRequest(CommandRequestWithChannel, ptz.GetZoomFocusRequest):
    """Get Zoom and Focus"""

    __slots__ = ()

    COMMAND: Final = "GetZoomFocus"

    def __init__(
        self,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id


_ZOOMFOCUS_KEY: Final = "ZoomFocus"


class GetZoomFocusResponse(CommandResponse, ptz.GetZoomFocusResponse):
    """Get Zoom/Focus Response"""

    __slots__ = ()

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetAutoFocusRequest.COMMAND)

    def _get_sub_value(self) -> dict:
        return (
            value.get(_ZOOMFOCUS_KEY, None)
            if (value := self._get_value()) is not None
            else None
        )

    channel_id = GetTatternResponse.channel_id

    @property
    def state(self):
        return ZoomFocus(self._get_sub_value())


_DEFAULT_ZOOMOPERATION: Final = typings.ZoomOperation.ZOOM
_DEFAULT_ZOOMOPERATION_STR: Final = ZOOMOPERATION_STR_MAP[_DEFAULT_ZOOMOPERATION]


class SetZoomFocusRequest(CommandRequest, ptz.SetZoomFocusRequest):
    """Set Zoom or Focus"""

    __slots__ = ()

    COMMAND: Final = "StartZoomFocus"

    def __init__(
        self,
        operation: typings.ZoomOperation,
        position: int,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id
        self.operation = operation
        self.position = position

    def _get_sub_value(self, create=False) -> dict:
        _key: Final = _ZOOMFOCUS_KEY
        if (parameter := self._get_parameter(create)) is None:
            return None
        if _key in parameter or not create:
            return parameter.get(_key, None)
        return parameter.setdefault(_key, {})

    @property
    def _sub_value(self):
        return self._get_sub_value(True)

    channel_id = SetTatternRequest.channel_id

    @property
    def operation(self) -> typings.ZoomOperation:
        if (value := self._get_sub_value()) is None:
            return _DEFAULT_ZOOMOPERATION
        return STR_ZOOMOPERATION_MAP[value.get("op", _DEFAULT_ZOOMOPERATION_STR)]

    @operation.setter
    def operation(self, value):
        self._sub_value["op"] = ZOOMOPERATION_STR_MAP[value or _DEFAULT_ZOOMOPERATION]

    @property
    def position(self) -> int:
        if (value := self._get_sub_value()) is None:
            return 0
        return value.get("pos", 0)

    @position.setter
    def position(self, value):
        self._sub_value["pos"] = value


_DEFAULT_OPERATION: Final = typings.Operation.AUTO
_DEFAULT_OPERATION_STR: Final = OPERATION_STR_MAP[_DEFAULT_OPERATION]


class SetControlRequest(CommandRequestWithChannel, ptz.SetControlRequest):
    """PTZ Control"""

    __slots__ = ()

    COMMAND: Final = "PtzCtrl"

    def __init__(
        self,
        operation: typings.Operation,
        preset_id: int = None,
        speed: int = None,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id
        self.operation = operation
        self.preset_id = preset_id
        self.speed = speed

    @property
    def operation(self):
        if (value := self._get_parameter()) is None:
            return _DEFAULT_OPERATION
        return STR_OPERATION_MAP[value.get("op", _DEFAULT_OPERATION_STR)]

    @operation.setter
    def operation(self, value):
        self._parameter["op"] = OPERATION_STR_MAP[value or _DEFAULT_OPERATION]

    @property
    def preset_id(self) -> int | None:
        if (value := self._get_parameter()) is None:
            return None
        return value.get("id", None)

    @preset_id.setter
    def preset_id(self, value):
        if value is None:
            if (_value := self._get_parameter()) is not None:
                del _value["id"]
            return
        self._parameter["id"] = value

    @property
    def speed(self) -> int | None:
        if (value := self._get_parameter()) is None:
            return None
        return value.get("speed", None)

    @speed.setter
    def speed(self, value):
        if value is None:
            if (_value := self._get_parameter()) is not None:
                del _value["speed"]
            return
        self._parameter["speed"] = value



"""Secuirty REST Commands"""

from typing import Final, TypeGuard
from async_reolink.api.commands import security

from . import CommandRequest, CommandResponseTypes, CommandResponse

from ..security.models import LoginToken

# pylint:disable=missing-function-docstring


class LoginRequest(CommandRequest, security.LoginRequest):
    """REST Login Request"""

    __slots__ = ()

    COMMAND: Final = "Login"

    def __init__(
        self,
        user_name: str,
        password: str,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ) -> None:
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.user_name = user_name
        self.password = password

    def _get_login(self, create=False) -> dict:
        _key: Final = "User"
        if (parameter := self._get_parameter(create)) is None:
            return None
        if _key in parameter or not create:
            return parameter.get(_key, None)
        return parameter.setdefault(_key, {})

    @property
    def _login(self):
        return self._get_login(True)

    @property
    def user_name(self) -> str:
        return (
            value.get("userName", "")
            if (value := self._get_login()) is not None
            else ""
        )

    @user_name.setter
    def user_name(self, value):
        self._login["userName"] = value

    @property
    def password(self) -> str:
        return (
            value.get("password", "")
            if (value := self._get_login()) is not None
            else ""
        )

    @password.setter
    def password(self, value):
        self._login["password"] = value


class LoginResponse(CommandResponse, security.LoginResponse, test="is_response"):
    """REST Login Response"""

    __slots__ = ()

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, LoginRequest.COMMAND)

    def _get_token(self) -> dict:
        return (
            value.get("Token", None)
            if (value := self._get_value()) is not None
            else None
        )

    @property
    def token(self):
        return LoginToken(self._get_token)


class LogoutRequest(CommandRequest, security.LogoutRequest):
    """REST Logut Request"""

    __slots__ = ()

    COMMAND: Final = "Logout"

    def __init__(
        self, response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY
    ) -> None:
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type



"""REST Security"""

from __future__ import annotations
from time import time
from typing import Sequence

from async_reolink.api.security import Security as BaseSecurity
from .. import connection

from ..commands import CommandRequest
from ..commands.security import LoginRequest, LoginResponse, LogoutRequest


class Security(BaseSecurity):
    """REST security mixin"""

    def __init__(self, *args, **kwargs) -> None:
        self.__token = ""
        super().__init__(*args, **kwargs)
        if isinstance(self, connection.Connection):
            self._force_get_callbacks.append(self.__force_get_login)
        self.__token_expires: float = 0
        self.__last_pwd_hash = 0

    def __force_get_login(
        self, url: str, _: dict[str, str], commands: Sequence[CommandRequest]
    ):
        if len(commands) > 1 or not isinstance(commands[0], LoginRequest):
            if self.__token:
                return (False, url + f"?token={self.__token}")
            return
        return url + f"?cmd={commands[0].command}"

    @property
    def _auth_token(self):
        return self.__token

    @property
    def is_authenticated(self) -> bool:
        # we use a 1s offest to give time for simple checks to do an operation
        return bool(self.__token) and self.authentication_timeout > 1

    @property
    def authentication_timeout(self):
        return self.__token_expires - time()

    @property
    def authentication_id(self):
        return self.__last_pwd_hash

    async def _prelogin(self, username: str):
        # keep hash of username so we can logout on new info provided
        pwd_hash = hash(username)
        if self.__last_pwd_hash != pwd_hash:
            await self.logout()
            self.__last_pwd_hash = pwd_hash

        if isinstance(self, connection.Connection):
            if not self.is_connected:
                return False
        return True

    async def _process_login(self, response: LoginResponse) -> bool:
        token = response.token

        self.__token = token.name
        self.__token_expires = time() + token.lease_time

        return True

    def _create_login_request(self, username: str, password: str):
        return LoginRequest(username, password)

    def _create_logout_request(self) -> LogoutRequest:
        return LogoutRequest()

    def _clear_login(self):
        self.__token = ""
        self.__token_expires = 0



"""REST Network Commands"""

from typing import TYPE_CHECKING, Callable, Final, Mapping, TypeGuard, cast
from async_reolink.api.commands import network

from async_reolink.api.typings import StreamTypes
from ..typings import STREAMTYPES_STR_MAP

from ..network import models

from ..commands import (
    _CHANNEL_KEY,
    CommandRequest,
    CommandResponseTypes,
    CommandResponse,
)


# pylint:disable=missing-function-docstring
# pylint: disable=too-few-public-methods


class GetLocalLinkRequest(CommandRequest, network.GetLocalLinkRequest):
    """REST Get Local Link Request"""

    __slots__ = ()

    COMMAND: Final = "GetLocalLink"

    def __init__(
        self, response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY
    ) -> None:
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type


class GetLocalLinkResponse(
    CommandResponse, network.GetLocalLinkResponse, test="is_response"
):
    """REST Get Local Link Response"""

    __slots__ = ()

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetLocalLinkRequest.COMMAND)

    def _local_link(self) -> dict:
        return (
            value.get("LocalLink", None)
            if (value := self._get_value()) is not None
            else None
        )

    @property
    def local_link(self):
        # we are not passing the factory here since this object is meant to be detachable
        return models.LinkInfo(self._local_link())


class GetChannelStatusRequest(CommandRequest, network.GetChannelStatusRequest):
    """REST Get Channel Status Request"""

    __slots__ = ()

    COMMAND: Final = "GetChannelstatus"

    def __init__(
        self, response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY
    ) -> None:
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type


class GetChannelStatusResponse(
    CommandResponse, network.GetChannelStatusResponse, test="is_response"
):
    """REST Get Channel Status Response"""

    __slots__ = ()

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetChannelStatusRequest.COMMAND)

    @property
    def channels(self):
        # we are not passing the factory here since this object is meant to be detachable
        return models.ChannelStatuses(self._get_value())


class GetNetworkPortsRequest(CommandRequest, network.GetNetworkPortsRequest):
    """REST Get Network Ports"""

    __slots__ = ()

    COMMAND: Final = "GetNetPort"

    def __init__(
        self, response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY
    ) -> None:
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type


class GetNetworkPortsResponse(
    CommandResponse, network.GetLocalLinkResponse, test="is_response"
):
    """REST Get Local Link Response"""

    __slots__ = ()

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetNetworkPortsRequest.COMMAND)

    def _get_ports(self) -> dict:
        return (
            value.get("NetPort", None)
            if (value := self._get_value()) is not None
            else None
        )

    @property
    def ports(self):
        # we are not passing the factory here since this object is meant to be detachable
        return models.NetworkPorts(self._get_ports())


class GetRTSPUrlsRequest(CommandRequest, network.GetRTSPUrlsRequest):
    """REST Get RTSP Urls Request"""

    __slots__ = ()

    COMMAND: Final = "GetRtspUrl"

    def __init__(
        self,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ) -> None:
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id


class _RTSPUrls(Mapping[StreamTypes, str]):
    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        self._factory = factory

    def __getitem__(self, __k: StreamTypes) -> str:
        if (value := self._factory()) is None:
            return None
        return value.get(STREAMTYPES_STR_MAP[__k] + "Stream", None)

    def __contains__(self, __o: StreamTypes) -> bool:
        if (value := self._factory()) is None:
            return False
        return STREAMTYPES_STR_MAP[__o] + "Stream" in value

    def __iter__(self):
        if (value := self._factory()) is None:
            return
        for _k, _v in STREAMTYPES_STR_MAP.items():
            if _v + "Stream" in value:
                yield _k

    def __len__(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return len(
            (True for _v in STREAMTYPES_STR_MAP.values() if _v + "Stream" in value)
        )


class GetRTSPUrlsResponse(
    CommandResponse, network.GetRTSPUrlsResponse, test="is_response"
):
    """REST Get RTSP Urls Response"""

    __slots__ = ()

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetRTSPUrlsRequest.COMMAND)

    def _get_urls(self) -> dict:
        return (
            value.get("rtspUrl", None)
            if (value := self._get_value()) is not None
            else None
        )

    @property
    def channel_id(self) -> int:
        return (
            value.get(_CHANNEL_KEY, 0) if (value := self._get_urls()) is not None else 0
        )

    @property
    def urls(self):
        # we are not passing the factory here since this object is meant to be detachable
        return _RTSPUrls(self._get_urls())


class GetP2PRequest(CommandRequest, network.GetP2PRequest):
    """REST Get Local Link Request"""

    __slots__ = ()

    COMMAND: Final = "GetP2p"

    def __init__(
        self, response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY
    ) -> None:
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type


class GetP2PResponse(CommandResponse, network.GetP2PResponse, test="is_response"):
    """REST Get P2P Response"""

    __slots__ = ()

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetP2PRequest.COMMAND)

    def _get_info(self) -> dict:
        return (
            value.get("P2p", None) if (value := self._get_value()) is not None else None
        )

    @property
    def info(self):
        # we are not passing the factory here since this object is meant to be detachable
        return models.P2PInfo(self._get_info())



"""Rest Typings"""

from types import MappingProxyType
from typing import Callable, Final, Protocol, TypeVar, overload
from typing_extensions import Self

from async_reolink.api.typings import StreamTypes

STR_STREAMTYPES_MAP: Final = MappingProxyType(
    {_e.name.lower(): _e for _e in StreamTypes}
)

STREAMTYPES_STR_MAP: Final = MappingProxyType(
    {_v: _k for _k, _v in STR_STREAMTYPES_MAP.items()}
)

_T = TypeVar("_T")


class FactoryValue(Protocol[_T]):
    """Value factory protocol"""

    def __call__(self, create=False) -> _T:
        ...


_VT = TypeVar("_VT")


class Property(Protocol[_T, _VT]):
    """Property Protocol"""

    fget: Callable[[_T], _VT]
    fset: Callable[[_T, _VT], None] | None
    fdel: Callable[[_T], None]

    @overload
    def __get__(self, obj: None, cls: type) -> Self:
        ...

    @overload
    def _get__(self, obj: _T, cls: type) -> _VT:
        ...

    def __get__(self, obj, cls) -> _T:
        ...

    def __set__(self, obj, value) -> None:
        ...



"""Security Models"""


# pylint: disable=too-few-public-methods
# pylint:disable=missing-function-docstring

from typing import Callable


class LoginToken:
    """Login Token"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        super().__init__()
        self._factory = factory

    @property
    def lease_time(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("leaseTime", 0)

    @property
    def name(self) -> str:
        if (value := self._factory()) is None:
            return ""
        return value.get("name", "")


class LoginTokenV2(LoginToken):
    """Login Token V2"""

    __slots__ = ()

    @property
    def check(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("checkBasic", 0)

    @property
    def count(self) -> int:
        if (value := self._factory()) is None:
            return 0
        return value.get("countTotal", 0)


class DigestInfo:
    """Digest Info"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        super().__init__()
        self._factory = factory

    @property
    def cnonce(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("Cnonce", "")

    @property
    def method(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("Method", "")

    @property
    def nc(self):  # pylint: disable=invalid-name
        if (value := self._factory()) is None:
            return ""
        return value.get("Nc", "")

    @property
    def nonce(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("Nonce", "")

    @property
    def qop(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("Qop", "")

    @property
    def realm(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("Realm", "")

    @property
    def response(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("Response", "")

    @property
    def uri(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("Uri", "")

    @property
    def user_name(self):
        if (value := self._factory()) is None:
            return ""
        return value.get("UserName", "")



"""REST Record Commands"""

from typing import Callable, Final, Sequence, TypeGuard, TypeVar
from async_reolink.api.commands import record
from async_reolink.api.record import typings

from ..record.models import MutableSearch, SearchStatus, File

from . import (
    _CHANNEL_KEY,
    CommandRequest,
    CommandRequestWithChannel,
    CommandResponse,
    CommandResponseTypes,
)

# pylint:disable=missing-function-docstring


class GetSnapshotRequest(CommandRequestWithChannel, record.GetSnapshotRequest):
    """REST Get Snaposhot Request"""

    COMMAND: Final = "Snap"

    def __init__(
        self,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.channel_id = channel_id

    @property
    def raw_parameter(self):
        return self._parameter


class SearchRecordingsRequest(CommandRequest, record.SearchRecordingsRequest):
    """REST Search Recordings Request"""

    COMMAND: Final = "Search"

    def __init__(
        self,
        search: typings.Search,
        channel_id: int = 0,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ):
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.search = search
        self.channel_id = channel_id

    def _get_search(self, create=False) -> dict:
        _key: Final = "Search"
        if (parameter := self._get_parameter(create)) is None:
            return None
        if _key in parameter or not create:
            return parameter.get(_key, None)
        return parameter.setdefault(_key, {})

    @property
    def _search(self):
        return self._get_search(True)

    @property
    def channel_id(self) -> int:
        if (value := self._get_search()) is None:
            return 0
        return value.get(_CHANNEL_KEY, 0)

    @channel_id.setter
    def channel_id(self, value):
        self._search[_CHANNEL_KEY] = value

    @property
    def search(self):
        return MutableSearch(self._get_search)

    @search.setter
    def search(self, value: typings.Search):
        if self._get_search() is None:
            if not isinstance(value, MutableSearch):
                value = MutableSearch(value)
            self._parameter["Search"] = value._factory(True)
        elif value is not None:
            search = self.search

            search.end = value.end
            search.start = value.start
            search.status_only = value.status_only
            search.stream_type = value.stream_type


_T = TypeVar("_T")


class _FactorySequence(Sequence[_T]):

    __slots__ = ("_factory", "_get_value")

    def __init__(
        self, getter: Callable[[], list], factory: Callable[[Callable[[], dict]], _T]
    ) -> None:
        super().__init__()
        self._get_value = getter
        self._factory = factory

    def _get_item(self, __k: int) -> dict:
        return value[__k] if (value := self._get_value()) is not None else None

    def __getitem__(self, __k: int):
        def _factory():
            return self._get_item(__k)

        return self._factory(_factory)

    def __len__(self):
        if (_list := self._get_value()) is None:
            return 0
        return len(_list)


class SearchRecordingsResponse(
    CommandResponse, record.SearchRecordingsResponse, test="is_response"
):
    """REST Search Results"""

    __slots__ = ()

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, SearchRecordingsRequest.COMMAND)

    def _get_sub_value(self) -> list:
        return (
            value.get("SearchResult", None)
            if (value := self._get_value()) is not None
            else None
        )

    @property
    def channel_id(self) -> int:
        if (value := self._get_sub_value()) is None:
            return 0
        return value.get(_CHANNEL_KEY, 0)

    def _get_status(self) -> list:
        return (
            value.get("Status", None)
            if (value := self._get_sub_value()) is not None
            else None
        )

    @property
    def status(self):
        return _FactorySequence(self._get_status, SearchStatus)

    def _get_file(self) -> list:
        return (
            value.get("File", None)
            if (value := self._get_sub_value()) is not None
            else None
        )

    @property
    def files(self):
        return _FactorySequence(self._get_file, File)



"""REST Record"""

from datetime import datetime
from typing import Sequence
from async_reolink.api.typings import StreamTypes
from async_reolink.api.commands import (
    CommandRequest,
)
from async_reolink.api.record import Record as BaseRecord, typings

from async_reolink.rest.record.models import MutableSearch

from ..commands import _COMMAND_KEY, record
from .seed import Seed


from .. import connection


class Record(BaseRecord):
    """REST Record Mixin"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if isinstance(self, connection.Connection):
            self._force_get_callbacks.append(self.__force_get_login)

    def __force_get_login(
        self, _: str, query: dict[str, str], commands: Sequence[CommandRequest]
    ):
        command = commands[0]
        if len(commands) > 1 or not isinstance(command, record.GetSnapshotRequest):
            return
        query[_COMMAND_KEY] = command.command
        query.update(command.raw_parameter)
        query["rs"] = str(Seed())
        return True

    def _create_get_snapshot_request(self, channel: int):
        return record.GetSnapshotRequest(channel)

    def _create_search_request(self, channel: int, search: typings.Search):
        return record.SearchRecordingsRequest(channel, search)

    def _create_search(
        self,
        start_time: datetime,
        end_time: datetime,
        only_status: bool,
        stream_type: StreamTypes,
    ):
        search = MutableSearch()
        search.start = start_time
        search.end = end_time
        search.status_only = only_status
        search.stream_type = stream_type
        return search



"""AI Typings"""

from types import MappingProxyType
from typing import Final
from async_reolink.api.ai.typings import AITypes

STR_AITYPES_MAP: Final = MappingProxyType(
    {"dog_cat" if _e == AITypes.PET else _e.name.lower(): _e for _e in AITypes}
)
AITYPES_STR_MAP: Final = MappingProxyType(
    {_v: _k for _k, _v in STR_AITYPES_MAP.items()}
)



"""Encoding REST models"""

from typing import Callable, Mapping
from async_reolink.api.typings import StreamTypes
from async_reolink.api.encoding import typings

from .typings import STREAMTYPES_STR_MAP

# pylint:disable=missing-function-docstring


class StreamEncodingInfo(typings.StreamEncodingInfo):
    """REST Stream Encoding Info"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        super().__init__()
        self._factory = factory

    @property
    def bit_rate(self) -> int:
        if (value := self._factory()) is not None:
            return value.get("bitRate", 0)
        return 0

    @property
    def frame_rate(self) -> int:
        if (value := self._factory()) is not None:
            return value.get("frameRate", 0)
        return 0

    @property
    def gop(self) -> int:
        if (value := self._factory()) is not None:
            return value.get("gop", 0)
        return 0

    @property
    def height(self) -> int:
        if (value := self._factory()) is not None:
            return value.get("height", 0)
        return 0

    @property
    def width(self) -> int:
        if (value := self._factory()) is not None:
            return value.get("width", 0)
        return 0

    @property
    def profile(self) -> int:
        if (value := self._factory()) is not None:
            return value.get("profile", "")
        return ""

    @property
    def size(self) -> int:
        if (value := self._factory()) is not None:
            return value.get("size", "")
        return ""

    @property
    def video_type(self) -> int:
        if (value := self._factory()) is not None:
            return value.get("video_type", "")
        return ""


class _StreamMapping(Mapping[StreamTypes, StreamEncodingInfo]):
    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        super().__init__()
        self._factory = factory

    def __getitem__(self, __k: StreamTypes):
        def _get():
            if (_dict := self._factory()) is not None:
                return _dict.get(STREAMTYPES_STR_MAP[__k], None)
            return None

        return StreamEncodingInfo(_get)

    def __iter__(self):
        if (_dict := self._factory()) is None:
            return
        for __k in StreamTypes:
            if STREAMTYPES_STR_MAP[__k] in _dict:
                yield __k

    def __len__(self):
        if (_dict := self._factory()) is None:
            return 0
        return len((__k for __k in StreamTypes if STREAMTYPES_STR_MAP[__k] in _dict))


class EncodingInfo(typings.EncodingInfo):
    """REST Encoding Info"""

    __slots__ = ("_factory",)

    def __init__(self, factory: Callable[[], dict]) -> None:
        super().__init__()
        self._factory = factory

    @property
    def audio(self) -> int:
        if (value := self._factory()) is not None:
            return value.get("audio", 0)
        return 0

    @property
    def stream(self):
        """stream"""

        return _StreamMapping(self._factory)



"""REST Network Mixin"""

from typing import Mapping, Sequence
from async_reolink.api import network
from async_reolink.api.led.typings import LightStates, LightingSchedule
from async_reolink.api.ai.typings import AITypes
from async_reolink.api.typings import PercentValue

from ..commands import network as commands


class Network(network.Network):
    """REST Network Mixin"""

    def _create_get_channel_status_request(self):
        return commands.GetChannelStatusRequest()

    def _create_get_local_link_request(self):
        return commands.GetLocalLinkRequest()

    def _create_get_p2p_request(self):
        return commands.GetP2PRequest()

    def _create_get_ports_request(self):
        return commands.GetNetworkPortsRequest()

    def _create_get_rtsp_urls_request(self, channel_id: int = 0):
        return commands.GetRTSPUrlsRequest(channel_id)



"""AI Mixin"""

from typing import Mapping

from async_reolink.api import ai
from async_reolink.api.ai.typings import AITypes

from ..commands import ai as commands


class AI(ai.AI):
    """Rest AI Mixin"""

    def _create_get_ai_state_request(self, channel: int):
        return commands.GetAiStateRequest(channel)

    def _create_get_ai_config_request(self, channel: int):
        return commands.GetAiStateRequest(channel)

    def _create_set_ai_config(
        self,
        channel: int,
        detect: AITypes | set[AITypes] | Mapping[AITypes, bool] | None,
        track: AITypes | set[AITypes] | Mapping[AITypes, bool] | None,
    ):
        return commands.SetAiConfigRequest(channel, detect, track)



"""REST LED Typings"""

from types import MappingProxyType
from typing import Final
from async_reolink.api.led.typings import LightStates

STR_LIGHTSTATES_MAP: Final = MappingProxyType(
    {"Auto": LightStates.AUTO, "On": LightStates.ON, "Off": LightStates.OFF}
)

LIGHTSTATES_STR_MAP: Final = MappingProxyType(
    {_v: _k for _k, _v in STR_LIGHTSTATES_MAP.items()}
)



"""System REST Commands"""

from typing import Final, TypeGuard
from async_reolink.api.commands import system

from ..system.models import DaylightSavingsTimeInfo, DeviceInfo, TimeInfo

from ..commands import (
    CommandRequest,
    CommandResponseTypes,
    CommandResponse,
)

from ..system.capabilities import Capabilities

# pylint:disable=missing-function-docstring
# pylint: disable=too-few-public-methods


class GetAbilitiesRequest(CommandRequest, system.GetAbilitiesRequest):
    """REST Get Capabilities Request"""

    COMMAND: Final = "GetAbility"

    def __init__(
        self,
        username: str,
        response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY,
    ) -> None:
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
        self.user_name = username

    def _get_user(self, create=False) -> dict:
        _key: Final = "User"
        if (parameter := self._get_parameter(create)) is None:
            return None
        if _key in parameter or not create:
            return parameter.get(_key, None)
        return parameter.setdefault(_key, {})

    @property
    def _user(self) -> dict:
        return self._get_user(True)

    @property
    def user_name(self) -> str:
        return (
            value.get("userName", "") if (value := self._get_user()) is not None else ""
        )

    @user_name.setter
    def user_name(self, value):
        self._user["userName"] = value


class GetAbilitiesResponse(
    CommandResponse, system.GetAbilitiesResponse, test="is_response"
):
    """REST Get Capability Response"""

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetAbilitiesRequest.COMMAND)

    def _get_ability(self) -> dict:
        return (
            value.get("Ability", None)
            if (value := self._get_value()) is not None
            else None
        )

    @property
    def capabilities(self):
        # we are not passing the factory here since this object is meant to be updatable
        return Capabilities(self._get_ability())


class GetDeviceInfoRequest(CommandRequest, system.GetDeviceInfoRequest):
    """REST Get Device Info Request"""

    COMMAND: Final = "GetDevInfo"

    def __init__(
        self, response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY
    ) -> None:
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type


class GetDeviceInfoResponse(
    CommandResponse, system.GetDeviceInfoResponse, test="is_response"
):
    """REST Get Device Info Response"""

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetDeviceInfoRequest.COMMAND)

    def _get_info(self) -> dict:
        return (
            value.get("DevInfo", None)
            if (value := self._get_value()) is not None
            else None
        )

    @property
    def info(self):
        """device info"""
        # we are not passing the factory here since this object is meant to be updatable
        return DeviceInfo(self._get_info())


class GetTimeRequest(CommandRequest, system.GetTimeRequest):
    """REST Get Time Request"""

    COMMAND: Final = "GetTime"

    def __init__(
        self, response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY
    ) -> None:
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type


class GetTimeResponse(CommandResponse, system.GetTimeResponse, test="is_response"):
    """REST Get Time Response"""

    @classmethod
    def is_response(cls, value: any, /):  # pylint: disable=signature-differs
        return super().is_response(value, GetTimeRequest.COMMAND)

    def _get_dst(self) -> dict:
        return (
            value.get("Dst", None) if (value := self._get_value()) is not None else None
        )

    @property
    def dst(self):
        return DaylightSavingsTimeInfo(self._get_dst)

    def _get_time(self) -> dict:
        return (
            value.get("Time", None)
            if (value := self._get_value()) is not None
            else None
        )

    @property
    def time(self):
        return TimeInfo(self._get_time)


class RebootRequest(CommandRequest, system.RebootRequest):
    """REST Reboot Request"""

    COMMAND: Final = "Reboot"

    def __init__(
        self, response_type: CommandResponseTypes = CommandResponseTypes.VALUE_ONLY
    ) -> None:
        super().__init__()
        self.command = type(self).COMMAND
        self.response_type = response_type
